# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import shutil

import pkg_resources.extern.packaging.version
import pytest
import yaml

from swh.core import config

pytest_v = pkg_resources.get_distribution("pytest").parsed_version
if pytest_v < pkg_resources.extern.packaging.version.parse("3.9"):

    @pytest.fixture
    def tmp_path():
        import pathlib
        import tempfile

        with tempfile.TemporaryDirectory() as tmpdir:
            yield pathlib.Path(tmpdir)


default_conf = {
    "a": ("int", 2),
    "b": ("string", "default-string"),
    "c": ("bool", True),
    "d": ("int", 10),
    "e": ("int", None),
    "f": ("bool", None),
    "g": ("string", None),
    "h": ("bool", True),
    "i": ("bool", True),
    "ls": ("list[str]", ["a", "b", "c"]),
    "li": ("list[int]", [42, 43]),
}

other_default_conf = {
    "a": ("int", 3),
}

full_default_conf = default_conf.copy()
full_default_conf["a"] = other_default_conf["a"]

parsed_default_conf = {key: value for key, (type, value) in default_conf.items()}

parsed_conffile = {
    "a": 1,
    "b": "this is a string",
    "c": True,
    "d": 10,
    "e": None,
    "f": None,
    "g": None,
    "h": False,
    "i": True,
    "ls": ["list", "of", "strings"],
    "li": [1, 2, 3, 4],
}


@pytest.fixture
def swh_config(tmp_path):
    # create a temporary folder
    conffile = tmp_path / "config.yml"
    conf_contents = """
a: 1
b: this is a string
c: true
h: false
ls: list, of, strings
li: 1, 2, 3, 4
"""
    conffile.open("w").write(conf_contents)
    return conffile


@pytest.fixture
def swh_config_unreadable(swh_config):
    # Create an unreadable, proper configuration file
    os.chmod(str(swh_config), 0o000)
    yield swh_config
    # Make the broken perms file readable again to be able to remove them
    os.chmod(str(swh_config), 0o644)


@pytest.fixture
def swh_config_unreadable_dir(swh_config):
    # Create a proper configuration file in an unreadable directory
    perms_broken_dir = swh_config.parent / "unreadabledir"
    perms_broken_dir.mkdir()
    shutil.move(str(swh_config), str(perms_broken_dir))
    os.chmod(str(perms_broken_dir), 0o000)
    yield perms_broken_dir / swh_config.name
    # Make the broken perms items readable again to be able to remove them
    os.chmod(str(perms_broken_dir), 0o755)


@pytest.fixture
def swh_config_empty(tmp_path):
    # create a temporary folder
    conffile = tmp_path / "config.yml"
    conffile.touch()
    return conffile


def test_read(swh_config):
    # when
    res = config.read(str(swh_config), default_conf)

    # then
    assert res == parsed_conffile


def test_read_no_default_conf(swh_config):
    """If no default config if provided to read, this should directly parse the config file
    yaml

    """
    config_path = str(swh_config)
    actual_config = config.read(config_path)
    with open(config_path) as f:
        expected_config = yaml.safe_load(f)
    assert actual_config == expected_config


def test_read_empty_file():
    # when
    res = config.read(None, default_conf)

    # then
    assert res == parsed_default_conf


def test_support_non_existing_conffile(tmp_path):
    # when
    res = config.read(str(tmp_path / "void.yml"), default_conf)

    # then
    assert res == parsed_default_conf


def test_support_empty_conffile(swh_config_empty):
    # when
    res = config.read(str(swh_config_empty), default_conf)

    # then
    assert res == parsed_default_conf


def test_raise_on_broken_directory_perms(swh_config_unreadable_dir):
    with pytest.raises(PermissionError):
        config.read(str(swh_config_unreadable_dir), default_conf)


def test_raise_on_broken_file_perms(swh_config_unreadable):
    with pytest.raises(PermissionError):
        config.read(str(swh_config_unreadable), default_conf)


def test_merge_default_configs():
    # when
    res = config.merge_default_configs(default_conf, other_default_conf)

    # then
    assert res == full_default_conf


def test_priority_read_nonexist_conf(swh_config):
    noexist = str(swh_config.parent / "void.yml")
    # when
    res = config.priority_read([noexist, str(swh_config)], default_conf)

    # then
    assert res == parsed_conffile


def test_priority_read_conf_nonexist_empty(swh_config):
    noexist = swh_config.parent / "void.yml"
    empty = swh_config.parent / "empty.yml"
    empty.touch()

    # when
    res = config.priority_read(
        [str(p) for p in (swh_config, noexist, empty)], default_conf
    )

    # then
    assert res == parsed_conffile


def test_priority_read_empty_conf_nonexist(swh_config):
    noexist = swh_config.parent / "void.yml"
    empty = swh_config.parent / "empty.yml"
    empty.touch()

    # when
    res = config.priority_read(
        [str(p) for p in (empty, swh_config, noexist)], default_conf
    )

    # then
    assert res == parsed_default_conf


def test_swh_config_paths():
    res = config.swh_config_paths("foo/bar.yml")

    assert res == [
        "~/.config/swh/foo/bar.yml",
        "~/.swh/foo/bar.yml",
        "/etc/softwareheritage/foo/bar.yml",
    ]


def test_prepare_folder(tmp_path):
    # given
    conf = {
        "path1": str(tmp_path / "path1"),
        "path2": str(tmp_path / "path2" / "depth1"),
    }

    # the folders does not exists
    assert not os.path.exists(conf["path1"]), "path1 should not exist."
    assert not os.path.exists(conf["path2"]), "path2 should not exist."

    # when
    config.prepare_folders(conf, "path1")

    # path1 exists but not path2
    assert os.path.exists(conf["path1"]), "path1 should now exist!"
    assert not os.path.exists(conf["path2"]), "path2 should not exist."

    # path1 already exists, skips it but creates path2
    config.prepare_folders(conf, "path1", "path2")

    assert os.path.exists(conf["path1"]), "path1 should still exist!"
    assert os.path.exists(conf["path2"]), "path2 should now exist."


def test_merge_config():
    cfg_a = {
        "a": 42,
        "b": [1, 2, 3],
        "c": None,
        "d": {"gheez": 27},
        "e": {
            "ea": "Mr. Bungle",
            "eb": None,
            "ec": [11, 12, 13],
            "ed": {"eda": "Secret Chief 3", "edb": "Faith No More"},
            "ee": 451,
        },
        "f": "Janis",
    }
    cfg_b = {
        "a": 43,
        "b": [41, 42, 43],
        "c": "Tom Waits",
        "d": None,
        "e": {
            "ea": "Igorrr",
            "ec": [51, 52],
            "ed": {"edb": "Sleepytime Gorilla Museum", "edc": "Nils Peter Molvaer"},
        },
        "g": "Hüsker Dü",
    }

    # merge A, B
    cfg_m = config.merge_configs(cfg_a, cfg_b)
    assert cfg_m == {
        "a": 43,  # b takes precedence
        "b": [41, 42, 43],  # b takes precedence
        "c": "Tom Waits",  # b takes precedence
        "d": None,  # b['d'] takes precedence (explicit None)
        "e": {
            "ea": "Igorrr",  # a takes precedence
            "eb": None,  # only in a
            "ec": [51, 52],  # b takes precedence
            "ed": {
                "eda": "Secret Chief 3",  # only in a
                "edb": "Sleepytime Gorilla Museum",  # b takes precedence
                "edc": "Nils Peter Molvaer",
            },  # only defined in b
            "ee": 451,
        },
        "f": "Janis",  # only defined in a
        "g": "Hüsker Dü",  # only defined in b
    }

    # merge B, A
    cfg_m = config.merge_configs(cfg_b, cfg_a)
    assert cfg_m == {
        "a": 42,  # a takes precedence
        "b": [1, 2, 3],  # a takes precedence
        "c": None,  # a takes precedence
        "d": {"gheez": 27},  # a takes precedence
        "e": {
            "ea": "Mr. Bungle",  # a takes precedence
            "eb": None,  # only defined in a
            "ec": [11, 12, 13],  # a takes precedence
            "ed": {
                "eda": "Secret Chief 3",  # only in a
                "edb": "Faith No More",  # a takes precedence
                "edc": "Nils Peter Molvaer",
            },  # only in b
            "ee": 451,
        },
        "f": "Janis",  # only in a
        "g": "Hüsker Dü",  # only in b
    }


def test_merge_config_type_error():
    for v in (1, "str", None):
        with pytest.raises(TypeError):
            config.merge_configs(v, {})
        with pytest.raises(TypeError):
            config.merge_configs({}, v)

    for v in (1, "str"):
        with pytest.raises(TypeError):
            config.merge_configs({"a": v}, {"a": {}})
        with pytest.raises(TypeError):
            config.merge_configs({"a": {}}, {"a": v})


def test_load_from_envvar_no_environment_var_swh_config_filename_set():
    """Without SWH_CONFIG_FILENAME set, load_from_envvar raises"""

    with pytest.raises(AssertionError, match="SWH_CONFIG_FILENAME environment"):
        config.load_from_envvar()


def test_load_from_envvar_no_default_config(swh_config, monkeypatch):
    config_path = str(swh_config)
    monkeypatch.setenv("SWH_CONFIG_FILENAME", config_path)

    actual_config = config.load_from_envvar()

    expected_config = config.read(config_path)
    assert actual_config == expected_config


def test_load_from_envvar_with_default_config(swh_config, monkeypatch):
    default_config = {
        "number": 666,
        "something-cool": ["something", "cool"],
    }

    config_path = str(swh_config)
    monkeypatch.setenv("SWH_CONFIG_FILENAME", config_path)

    actual_config = config.load_from_envvar(default_config)

    expected_config = config.read(config_path)
    expected_config.update(
        {
            "number": 666,
            "something-cool": ["something", "cool"],
        }
    )

    assert actual_config == expected_config



# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os
from typing import Dict, List, Optional

import pkg_resources

logger = logging.getLogger(__name__)


def get_sentry_release():
    main_package = os.environ.get("SWH_MAIN_PACKAGE")
    if main_package:
        version = pkg_resources.get_distribution(main_package).version
        return f"{main_package}@{version}"
    else:
        return None


def override_with_bool_envvar(envvar: str, default: bool) -> bool:
    """Override the `default` with the environment variable `envvar` parsed as a boolean"""
    envvalue = os.environ.get(envvar, "")
    if envvalue.lower() in ("t", "true", "y", "yes", "1"):
        return True
    elif envvalue.lower() in ("f", "false", "n", "no", "0"):
        return False
    else:
        if envvalue:
            logger.warning(
                "Could not interpret environment variable %s=%r as boolean, "
                "using default value %s",
                envvar,
                envvalue,
                default,
            )
        return default


def init_sentry(
    sentry_dsn: Optional[str] = None,
    *,
    environment: Optional[str] = None,
    debug: bool = False,
    disable_logging_events: bool = False,
    integrations: Optional[List] = None,
    extra_kwargs: Optional[Dict] = None,
):
    """Configure the sentry integration

    Args:
      sentry_dsn: Sentry DSN; where sentry report will be sent. Overridden by
        :envvar:`SWH_SENTRY_DSN`
      environment: Sentry environment. Overridden by :envvar:`SWH_SENTRY_ENVIRONMENT`
      debug: turn on Sentry SDK debug mode. Overridden by :envvar:`SWH_SENTRY_DEBUG`
      disable_logging_events: if set, disable the automatic reporting of error/exception
        log entries as Sentry events. Overridden by
        :envvar:`SWH_SENTRY_DISABLE_LOGGING_EVENTS`
      integrations: list of dedicated Sentry integrations to include
      extra_kwargs: dict of additional parameters passed to :func:`sentry_sdk.init`

    """
    if integrations is None:
        integrations = []
    if extra_kwargs is None:
        extra_kwargs = {}

    sentry_dsn = os.environ.get("SWH_SENTRY_DSN", sentry_dsn)
    environment = os.environ.get("SWH_SENTRY_ENVIRONMENT", environment)

    debug = override_with_bool_envvar("SWH_SENTRY_DEBUG", debug)
    disable_logging_events = override_with_bool_envvar(
        "SWH_SENTRY_DISABLE_LOGGING_EVENTS", disable_logging_events
    )

    if sentry_dsn:
        import sentry_sdk

        if disable_logging_events:
            from sentry_sdk.integrations.logging import LoggingIntegration

            integrations.append(LoggingIntegration(event_level=None))

        sentry_sdk.init(
            release=get_sentry_release(),
            environment=environment,
            dsn=sentry_dsn,
            integrations=integrations,
            debug=debug,
            **extra_kwargs,
        )



# Copyright (C) 2015-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.core import utils


def test_grouper():
    # given
    actual_data = utils.grouper((i for i in range(0, 9)), 2)

    out = []
    for d in actual_data:
        out.append(list(d))  # force generator resolution for checks

    assert out == [[0, 1], [2, 3], [4, 5], [6, 7], [8]]

    # given
    actual_data = utils.grouper((i for i in range(9, 0, -1)), 4)

    out = []
    for d in actual_data:
        out.append(list(d))  # force generator resolution for checks

    assert out == [[9, 8, 7, 6], [5, 4, 3, 2], [1]]


def test_grouper_with_stop_value():
    # given
    actual_data = utils.grouper(((i, i + 1) for i in range(0, 9)), 2)

    out = []
    for d in actual_data:
        out.append(list(d))  # force generator resolution for checks

    assert out == [
        [(0, 1), (1, 2)],
        [(2, 3), (3, 4)],
        [(4, 5), (5, 6)],
        [(6, 7), (7, 8)],
        [(8, 9)],
    ]

    # given
    actual_data = utils.grouper((i for i in range(9, 0, -1)), 4)

    out = []
    for d in actual_data:
        out.append(list(d))  # force generator resolution for checks

    assert out == [[9, 8, 7, 6], [5, 4, 3, 2], [1]]


def test_iter_chunks():
    def chunks(input_, remainder):
        return list(utils.iter_chunks(input_, 3, remainder=remainder))

    # all even, remainder=False
    assert chunks(["ab", "cd", "ef"], False) == ["abc", "def"]
    assert chunks(["abc", "def"], False) == ["abc", "def"]
    assert chunks(["abcd", "ef"], False) == ["abc", "def"]

    # all even, remainder=True
    assert chunks(["ab", "cd", "ef"], True) == ["abc", "def"]
    assert chunks(["abc", "def"], True) == ["abc", "def"]
    assert chunks(["abcd", "ef"], True) == ["abc", "def"]

    # uneven, remainder=False
    assert chunks([], False) == []
    assert chunks(["ab"], False) == []
    assert chunks(["ab", "cd", "ef", "g"], False) == ["abc", "def"]
    assert chunks(["ab", "cd", "efg"], False) == ["abc", "def"]
    assert chunks(["abc", "def", "g"], False) == ["abc", "def"]
    assert chunks(["abcd", "ef", "g"], False) == ["abc", "def"]

    # uneven, remainder=True
    assert chunks([], True) == []
    assert chunks(["ab"], True) == ["ab"]
    assert chunks(["ab", "cd", "ef", "g"], True) == ["abc", "def", "g"]
    assert chunks(["ab", "cd", "efg"], True) == ["abc", "def", "g"]
    assert chunks(["abc", "def", "g"], True) == ["abc", "def", "g"]
    assert chunks(["abcd", "ef", "g"], True) == ["abc", "def", "g"]


def test_backslashescape_errors():
    raw_data_err = b"abcd\x80"
    with pytest.raises(UnicodeDecodeError):
        raw_data_err.decode("utf-8", "strict")

    assert raw_data_err.decode("utf-8", "backslashescape") == "abcd\\x80"

    raw_data_ok = b"abcd\xc3\xa9"
    assert raw_data_ok.decode("utf-8", "backslashescape") == raw_data_ok.decode(
        "utf-8", "strict"
    )

    unicode_data = "abcdef\u00a3"
    assert unicode_data.encode("ascii", "backslashescape") == b"abcdef\\xa3"


def test_encode_with_unescape():
    valid_data = "\\x01020304\\x00"
    valid_data_encoded = b"\x01020304\x00"

    assert valid_data_encoded == utils.encode_with_unescape(valid_data)


def test_encode_with_unescape_invalid_escape():
    invalid_data = "test\\abcd"

    with pytest.raises(ValueError) as exc:
        utils.encode_with_unescape(invalid_data)

    assert "invalid escape" in exc.value.args[0]
    assert "position 4" in exc.value.args[0]


def test_decode_with_escape():
    backslashes = b"foo\\bar\\\\baz"
    backslashes_escaped = "foo\\\\bar\\\\\\\\baz"

    assert backslashes_escaped == utils.decode_with_escape(backslashes)

    valid_utf8 = b"foo\xc3\xa2"
    valid_utf8_escaped = "foo\u00e2"

    assert valid_utf8_escaped == utils.decode_with_escape(valid_utf8)

    invalid_utf8 = b"foo\xa2"
    invalid_utf8_escaped = "foo\\xa2"

    assert invalid_utf8_escaped == utils.decode_with_escape(invalid_utf8)

    valid_utf8_nul = b"foo\xc3\xa2\x00"
    valid_utf8_nul_escaped = "foo\u00e2\\x00"

    assert valid_utf8_nul_escaped == utils.decode_with_escape(valid_utf8_nul)


def test_commonname():
    # when
    actual_commonname = utils.commonname("/some/where/to/", "/some/where/to/go/to")
    # then
    assert "go/to" == actual_commonname

    # when
    actual_commonname2 = utils.commonname(b"/some/where/to/", b"/some/where/to/go/to")
    # then
    assert b"go/to" == actual_commonname2


def test_numfile_sotkey():
    assert utils.numfile_sortkey("00-xxx.sql") == (0, "-xxx.sql")
    assert utils.numfile_sortkey("01-xxx.sql") == (1, "-xxx.sql")
    assert utils.numfile_sortkey("10-xxx.sql") == (10, "-xxx.sql")
    assert utils.numfile_sortkey("99-xxx.sql") == (99, "-xxx.sql")
    assert utils.numfile_sortkey("100-xxx.sql") == (100, "-xxx.sql")
    assert utils.numfile_sortkey("00100-xxx.sql") == (100, "-xxx.sql")
    assert utils.numfile_sortkey("1.sql") == (1, ".sql")
    assert utils.numfile_sortkey("1") == (1, "")
    assert utils.numfile_sortkey("toto-01.sql") == (999999, "toto-01.sql")


def test_basename_sotkey():
    assert utils.basename_sortkey("00-xxx.sql") == (0, "-xxx.sql")
    assert utils.basename_sortkey("path/to/00-xxx.sql") == (0, "-xxx.sql")



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import glob
from importlib import import_module
import logging
import subprocess
from typing import Callable, Iterable, Iterator, List, Optional, Sequence, Set, Union
import warnings

from _pytest.fixtures import FixtureRequest
from deprecated import deprecated
import psycopg2
import pytest
from pytest_postgresql.compat import check_for_psycopg2, connection
from pytest_postgresql.executor import PostgreSQLExecutor
from pytest_postgresql.executor_noop import NoopExecutor
from pytest_postgresql.janitor import DatabaseJanitor

from swh.core.db.db_utils import initialize_database_for_module
from swh.core.utils import basename_sortkey

# to keep mypy happy regardless pytest-postgresql version
try:
    _pytest_pgsql_get_config_module = import_module("pytest_postgresql.config")
except ImportError:
    # pytest_postgresql < 3.0.0
    _pytest_pgsql_get_config_module = import_module("pytest_postgresql.factories")

_pytest_postgresql_get_config = getattr(_pytest_pgsql_get_config_module, "get_config")


logger = logging.getLogger(__name__)

initialize_database_for_module = deprecated(
    version="2.10",
    reason="Use swh.core.db.db_utils.initialize_database_for_module instead.",
)(initialize_database_for_module)

warnings.warn(
    "This pytest plugin is deprecated, it should not be used any more.",
    category=DeprecationWarning,
)


class SWHDatabaseJanitor(DatabaseJanitor):
    """SWH database janitor implementation with a a different setup/teardown policy than
    than the stock one. Instead of dropping, creating and initializing the database for
    each test, it creates and initializes the db once, then truncates the tables (and
    sequences) in between tests.

    This is needed to have acceptable test performances.

    """

    def __init__(
        self,
        user: str,
        host: str,
        port: int,
        dbname: str,
        version: Union[str, float],
        password: Optional[str] = None,
        isolation_level: Optional[int] = None,
        connection_timeout: int = 60,
        dump_files: Optional[Union[str, Sequence[str]]] = None,
        no_truncate_tables: Set[str] = set(),
        no_db_drop: bool = False,
    ) -> None:
        super().__init__(user, host, port, dbname, version)
        # do no truncate the following tables
        self.no_truncate_tables = set(no_truncate_tables)
        self.no_db_drop = no_db_drop
        self.dump_files = dump_files

    def psql_exec(self, fname: str) -> None:
        conninfo = (
            f"host={self.host} user={self.user} port={self.port} dbname={self.dbname}"
        )

        subprocess.check_call(
            [
                "psql",
                "--quiet",
                "--no-psqlrc",
                "-v",
                "ON_ERROR_STOP=1",
                "-d",
                conninfo,
                "-f",
                fname,
            ]
        )

    def db_reset(self) -> None:
        """Truncate tables (all but self.no_truncate_tables set) and sequences"""
        with psycopg2.connect(
            dbname=self.dbname,
            user=self.user,
            host=self.host,
            port=self.port,
        ) as cnx:
            with cnx.cursor() as cur:
                cur.execute(
                    "SELECT table_name FROM information_schema.tables "
                    "WHERE table_schema = %s",
                    ("public",),
                )
                all_tables = set(table for (table,) in cur.fetchall())
                tables_to_truncate = all_tables - self.no_truncate_tables

                for table in tables_to_truncate:
                    cur.execute("TRUNCATE TABLE %s CASCADE" % table)

                cur.execute(
                    "SELECT sequence_name FROM information_schema.sequences "
                    "WHERE sequence_schema = %s",
                    ("public",),
                )
                seqs = set(seq for (seq,) in cur.fetchall())
                for seq in seqs:
                    cur.execute("ALTER SEQUENCE %s RESTART;" % seq)
            cnx.commit()

    def _db_exists(self, cur, dbname):
        cur.execute(
            "SELECT EXISTS "
            "(SELECT datname FROM pg_catalog.pg_database WHERE datname= %s);",
            (dbname,),
        )
        row = cur.fetchone()
        return (row is not None) and row[0]

    def init(self) -> None:
        """Create database in postgresql out of a template it if it exists, bare
        creation otherwise."""
        template_name = f"{self.dbname}_tmpl"
        logger.debug("Initialize DB %s", self.dbname)
        with self.cursor() as cur:
            tmpl_exists = self._db_exists(cur, template_name)
            db_exists = self._db_exists(cur, self.dbname)
            if not db_exists:
                if tmpl_exists:
                    logger.debug(
                        "Create %s from template %s", self.dbname, template_name
                    )
                    cur.execute(
                        f'CREATE DATABASE "{self.dbname}" TEMPLATE "{template_name}";'
                    )
                else:
                    logger.debug("Create %s from scratch", self.dbname)
                    cur.execute(f'CREATE DATABASE "{self.dbname}";')
                if self.dump_files:
                    logger.warning(
                        "Using dump_files on the postgresql_fact fixture "
                        "is deprecated. See swh.core documentation for more "
                        "details."
                    )
                    for dump_file in gen_dump_files(self.dump_files):
                        logger.info(f"Loading {dump_file}")
                        self.psql_exec(dump_file)
            else:
                logger.debug("Reset %s", self.dbname)
                self.db_reset()

    def drop(self) -> None:
        """Drop database in postgresql."""
        if self.no_db_drop:
            with self.cursor() as cur:
                self._terminate_connection(cur, self.dbname)
        else:
            super().drop()


# the postgres_fact factory fixture below is mostly a copy of the code
# from pytest-postgresql. We need a custom version here to be able to
# specify our version of the DBJanitor we use.
@deprecated(version="2.10", reason="Use stock pytest_postgresql factory instead")
def postgresql_fact(
    process_fixture_name: str,
    dbname: Optional[str] = None,
    load: Optional[Sequence[Union[Callable, str]]] = None,
    isolation_level: Optional[int] = None,
    modname: Optional[str] = None,
    dump_files: Optional[Union[str, List[str]]] = None,
    no_truncate_tables: Set[str] = {"dbversion"},
    no_db_drop: bool = False,
) -> Callable[[FixtureRequest], Iterator[connection]]:
    """
    Return connection fixture factory for PostgreSQL.

    :param process_fixture_name: name of the process fixture
    :param dbname: database name
    :param load: SQL, function or function import paths to automatically load
                 into our test database
    :param isolation_level: optional postgresql isolation level
                            defaults to server's default
    :param modname: (swh) module name for which the database is created
    :dump_files: (deprecated, use load instead) list of sql script files to
                 execute after the database has been created
    :no_truncate_tables: list of table not to truncate between tests (only used
                         when no_db_drop is True)
    :no_db_drop: if True, keep the database between tests; in which case, the
                 database is reset (see SWHDatabaseJanitor.db_reset()) by truncating
                 most of the tables. Note that this makes de facto tests (potentially)
                 interdependent, use with extra caution.
    :returns: function which makes a connection to postgresql
    """

    @pytest.fixture
    def postgresql_factory(request: FixtureRequest) -> Iterator[connection]:
        """
        Fixture factory for PostgreSQL.

        :param request: fixture request object
        :returns: postgresql client
        """
        check_for_psycopg2()
        proc_fixture: Union[PostgreSQLExecutor, NoopExecutor] = request.getfixturevalue(
            process_fixture_name
        )

        pg_host = proc_fixture.host
        pg_port = proc_fixture.port
        pg_user = proc_fixture.user
        pg_password = proc_fixture.password
        pg_options = proc_fixture.options
        pg_db = dbname or proc_fixture.dbname
        pg_load = load or []
        assert pg_db is not None

        with SWHDatabaseJanitor(
            pg_user,
            pg_host,
            pg_port,
            pg_db,
            proc_fixture.version,
            pg_password,
            isolation_level=isolation_level,
            dump_files=dump_files,
            no_truncate_tables=no_truncate_tables,
            no_db_drop=no_db_drop,
        ) as janitor:
            db_connection: connection = psycopg2.connect(
                dbname=pg_db,
                user=pg_user,
                password=pg_password,
                host=pg_host,
                port=pg_port,
                options=pg_options,
            )
            for load_element in pg_load:
                janitor.load(load_element)
            try:
                yield db_connection
            finally:
                db_connection.close()

    return postgresql_factory


def gen_dump_files(dump_files: Union[str, Iterable[str]]) -> Iterator[str]:
    """Generate files potentially resolving glob patterns if any"""
    if isinstance(dump_files, str):
        dump_files = [dump_files]
    for dump_file in dump_files:
        if glob.has_magic(dump_file):
            # if the dump_file is a glob pattern one, resolve it
            yield from (
                fname for fname in sorted(glob.glob(dump_file), key=basename_sortkey)
            )
        else:
            # otherwise, just return the filename
            yield dump_file



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import time
from swh.core.github.pytest_plugin import fake_time_sleep, fake_time_time


@pytest.mark.parametrize("duration", [10, 20, -1])
def test_fake_time_sleep(duration):

    if duration < 0:
        with pytest.raises(ValueError, match="negative"):
            fake_time_sleep(duration, [])
    else:
        sleep_calls = []
        fake_time_sleep(duration, sleep_calls)
        assert duration in sleep_calls


def test_fake_time_time():
    assert fake_time_time() == 0


def test_monkeypatch_sleep_calls(monkeypatch_sleep_calls):

    sleeps = [10, 20, 30]
    for sleep in sleeps:
        # This adds the sleep number inside the monkeypatch_sleep_calls fixture
        time.sleep(sleep)
        assert sleep in monkeypatch_sleep_calls

    assert len(monkeypatch_sleep_calls) == len(sleeps)
    # This mocks time but adds nothing to the same fixture
    time.time()
    assert len(monkeypatch_sleep_calls) == len(sleeps)


def test_num_before_ratelimit(num_before_ratelimit):
    assert num_before_ratelimit == 0


def test_ratelimit_reset(ratelimit_reset):
    assert ratelimit_reset is None


def test_num_ratelimit(num_ratelimit):
    assert num_ratelimit is None



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import textwrap
from typing import List
from unittest.mock import patch

import click
from click.testing import CliRunner
import pkg_resources
import pytest
import yaml

help_msg_snippets = (
    (
        "Usage",
        (
            "swh [OPTIONS] COMMAND [ARGS]...",
            "Command line interface for Software Heritage.",
        ),
    ),
    (
        "Options",
        (
            "-l, --log-level",
            "--log-config",
            "--sentry-dsn",
            "-h, --help",
        ),
    ),
)


def get_section(cli_output: str, section: str) -> List[str]:
    """Get the given `section` of the `cli_output`"""
    result = []
    in_section = False
    for line in cli_output.splitlines():
        if not line:
            continue

        if in_section:
            if not line.startswith(" "):
                break
        else:
            if line.startswith(section):
                in_section = True

        if in_section:
            result.append(line)

    return result


def assert_section_contains(cli_output: str, section: str, snippet: str) -> bool:
    """Check that a given `section` of the `cli_output` contains the given `snippet`"""
    section_lines = get_section(cli_output, section)
    assert section_lines, "Section %s not found in output %r" % (section, cli_output)

    for line in section_lines:
        if snippet in line:
            return True
    else:
        assert False, "%r not found in section %r of output %r" % (
            snippet,
            section,
            cli_output,
        )


@pytest.fixture(autouse=True)
def reset_root_loglevel():
    root_level = logging.root.level
    yield
    logging.root.setLevel(root_level)


@pytest.fixture
def patched_dictconfig(mocker):
    yield mocker.patch("logging.config.dictConfig", autospec=True)


def test_swh_help(swhmain):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["-h"])
    assert result.exit_code == 0
    for section, snippets in help_msg_snippets:
        for snippet in snippets:
            assert_section_contains(result.output, section, snippet)

    result = runner.invoke(swhmain, ["--help"])
    assert result.exit_code == 0
    for section, snippets in help_msg_snippets:
        for snippet in snippets:
            assert_section_contains(result.output, section, snippet)


def test_command(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        result = runner.invoke(swhmain, ["test"])
    sentry_sdk_init.assert_not_called()
    assert result.exit_code == 0
    assert result.output.strip() == "Hello SWH!"


def test_loglevel_default(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 20
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""


def test_loglevel_error(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 40
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["-l", "ERROR", "test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""


def test_loglevel_debug(caplog, swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 10
        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(swhmain, ["-l", "DEBUG", "test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""


def test_sentry(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        result = runner.invoke(swhmain, ["--sentry-dsn", "test_dsn", "test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=False,
        integrations=[],
        release=None,
        environment=None,
    )


def test_sentry_debug(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        result = runner.invoke(
            swhmain, ["--sentry-dsn", "test_dsn", "--sentry-debug", "test"]
        )
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=True,
        integrations=[],
        release=None,
        environment=None,
    )


def test_sentry_env(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        env = {
            "SWH_SENTRY_DSN": "test_dsn",
            "SWH_SENTRY_DEBUG": "1",
        }
        result = runner.invoke(swhmain, ["test"], env=env, auto_envvar_prefix="SWH")
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=True,
        integrations=[],
        release=None,
        environment=None,
    )


def test_sentry_env_main_package(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        click.echo("Hello SWH!")

    runner = CliRunner()
    with patch("sentry_sdk.init") as sentry_sdk_init:
        env = {
            "SWH_SENTRY_DSN": "test_dsn",
            "SWH_MAIN_PACKAGE": "swh.core",
            "SWH_SENTRY_ENVIRONMENT": "tests",
        }
        result = runner.invoke(swhmain, ["test"], env=env, auto_envvar_prefix="SWH")
    assert result.exit_code == 0

    version = pkg_resources.get_distribution("swh.core").version

    assert result.output.strip() == """Hello SWH!"""
    sentry_sdk_init.assert_called_once_with(
        dsn="test_dsn",
        debug=False,
        integrations=[],
        release="swh.core@" + version,
        environment="tests",
    )


@pytest.fixture
def log_config_path(tmp_path):
    log_config = textwrap.dedent(
        """\
    ---
    version: 1
    formatters:
      formatter:
        format: 'custom format:%(name)s:%(levelname)s:%(message)s'
    handlers:
      console:
        class: logging.StreamHandler
        stream: ext://sys.stdout
        formatter: formatter
        level: DEBUG
    root:
      level: DEBUG
      handlers:
        - console
    loggers:
      dontshowdebug:
        level: INFO
    """
    )

    (tmp_path / "log_config.yml").write_text(log_config)

    yield str(tmp_path / "log_config.yml")


def test_log_config(log_config_path, swhmain, patched_dictconfig):
    """Check that --log-config properly calls :func:`logging.config.dictConfig`."""

    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        patched_dictconfig.assert_called_once_with(
            yaml.safe_load(open(log_config_path, "r"))
        )

        click.echo("Hello SWH!")

    runner = CliRunner()
    result = runner.invoke(
        swhmain,
        [
            "--log-config",
            log_config_path,
            "test",
        ],
    )

    assert result.exit_code == 0
    assert result.output.strip() == "Hello SWH!"


def test_log_config_log_level_interaction(log_config_path, swhmain, patched_dictconfig):
    """Check that --log-config and --log-level work properly together, by calling
    :func:`logging.config.dictConfig` then setting the loglevel of the root logger."""

    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == logging.DEBUG
        patched_dictconfig.assert_called_once_with(
            yaml.safe_load(open(log_config_path, "r"))
        )

        click.echo("Hello SWH!")

    assert logging.root.level != logging.DEBUG

    runner = CliRunner()
    result = runner.invoke(
        swhmain,
        [
            "--log-config",
            log_config_path,
            "--log-level",
            "DEBUG",
            "test",
        ],
    )

    assert result.exit_code == 0
    assert result.output.strip() == "Hello SWH!"


def test_multiple_log_level_behavior(swhmain):
    @swhmain.command(name="test")
    @click.pass_context
    def swhtest(ctx):
        assert logging.getLevelName(logging.root.level) == "DEBUG"
        assert logging.getLevelName(logging.getLogger("dontshowdebug").level) == "INFO"
        return 0

    runner = CliRunner()
    result = runner.invoke(
        swhmain,
        [
            "--log-level",
            "DEBUG",
            "--log-level",
            "dontshowdebug:INFO",
            "test",
        ],
    )

    assert result.exit_code == 0, result.output


def test_invalid_log_level(swhmain):
    runner = CliRunner()
    result = runner.invoke(swhmain, ["--log-level", "broken:broken:DEBUG"])

    assert result.exit_code != 0
    assert "Invalid log level specification" in result.output

    runner = CliRunner()
    result = runner.invoke(swhmain, ["--log-level", "UNKNOWN"])

    assert result.exit_code != 0
    assert "Log level UNKNOWN unknown" in result.output


def test_aliased_command(swhmain):
    @swhmain.command(name="canonical-test")
    @click.pass_context
    def swhtest(ctx):
        "A test command."
        click.echo("Hello SWH!")

    swhmain.add_alias(swhtest, "othername")

    runner = CliRunner()

    # check we have only 'canonical-test' listed in the usage help msg
    result = runner.invoke(swhmain, ["-h"])
    assert result.exit_code == 0
    assert "canonical-test  A test command." in result.output
    assert "othername" not in result.output

    # check we can execute the cmd with 'canonical-test'
    result = runner.invoke(swhmain, ["canonical-test"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""

    # check we can also execute the cmd with the alias 'othername'
    result = runner.invoke(swhmain, ["othername"])
    assert result.exit_code == 0
    assert result.output.strip() == """Hello SWH!"""



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import requests

from .conftest import DATADIR

# In this arborescence, we override in the local conftest.py module the
# "datadir" fixture to specify where to retrieve the data files from.


def test_requests_mock_datadir_with_datadir_fixture_override(requests_mock_datadir):
    """Override datadir fixture should retrieve data from elsewhere"""
    response = requests.get("https://example.com/file.json")
    assert response.ok
    assert response.json() == {"welcome": "you"}


def test_data_dir_override(datadir):
    assert datadir == DATADIR



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import glob
import os

from pytest_postgresql import factories

from swh.core.db import BaseDb
from swh.core.db.pytest_plugin import gen_dump_files, postgresql_fact

SQL_DIR = os.path.join(os.path.dirname(__file__), "data")

test_postgresql_proc = factories.postgresql_proc(
    dbname="fun",
    load=sorted(glob.glob(f"{SQL_DIR}/*.sql")),  # type: ignore[arg-type]
    # type ignored because load is typed as Optional[List[...]] instead of an
    # Optional[Sequence[...]] in pytest_postgresql<4
)

# db with special policy for tables dbversion and people
postgres_fun = postgresql_fact(
    "test_postgresql_proc",
    no_db_drop=True,
    no_truncate_tables={"dbversion", "people"},
)

postgres_fun2 = postgresql_fact(
    "test_postgresql_proc",
    dbname="fun2",
    load=sorted(glob.glob(f"{SQL_DIR}/*.sql")),
    no_truncate_tables={"dbversion", "people"},
    no_db_drop=True,
)


def test_smoke_test_fun_db_is_up(postgres_fun):
    """This ensures the db is created and configured according to its dumps files."""
    with BaseDb.connect(postgres_fun.dsn).cursor() as cur:
        cur.execute("select count(*) from dbversion")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 5

        cur.execute("select count(*) from fun")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 3

        cur.execute("select count(*) from people")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 2

        # in data, we requested a value already so it starts at 2
        cur.execute("select nextval('serial')")
        val = cur.fetchone()[0]
        assert val == 2


def test_smoke_test_fun2_db_is_up(postgres_fun2):
    """This ensures the db is created and configured according to its dumps files."""
    with BaseDb.connect(postgres_fun2.dsn).cursor() as cur:
        cur.execute("select count(*) from dbversion")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 5

        cur.execute("select count(*) from fun")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 3

        cur.execute("select count(*) from people")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 2

        # in data, we requested a value already so it starts at 2
        cur.execute("select nextval('serial')")
        val = cur.fetchone()[0]
        assert val == 2


def test_smoke_test_fun_db_is_still_up_and_got_reset(postgres_fun):
    """This ensures that within another tests, the 'fun' db is still up, created (and not
    configured again). This time, most of the data has been reset:
    - except for tables 'dbversion' and 'people' which were left as is
    - the other tables from the schema (here only "fun") got truncated
    - the sequences got truncated as well

    """
    with BaseDb.connect(postgres_fun.dsn).cursor() as cur:
        # db version is excluded from the truncate
        cur.execute("select count(*) from dbversion")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 5

        # people is also allowed not to be truncated
        cur.execute("select count(*) from people")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 2

        # table and sequence are reset
        cur.execute("select count(*) from fun")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 0

        cur.execute("select nextval('serial')")
        val = cur.fetchone()[0]
        assert val == 1


# db with no special policy for tables truncation, all tables are reset
postgres_people = postgresql_fact(
    "postgresql_proc",
    dbname="people",
    dump_files=f"{SQL_DIR}/*.sql",
    no_truncate_tables=set(),
    no_db_drop=True,
)


def test_gen_dump_files():
    files = [os.path.basename(fn) for fn in gen_dump_files(f"{SQL_DIR}/*.sql")]
    assert files == ["0-schema.sql", "1-data.sql"]


def test_smoke_test_people_db_up(postgres_people):
    """'people' db is up and configured"""
    with BaseDb.connect(postgres_people.dsn).cursor() as cur:
        cur.execute("select count(*) from dbversion")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 5

        cur.execute("select count(*) from people")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 2

        cur.execute("select count(*) from fun")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 3

        cur.execute("select nextval('serial')")
        val = cur.fetchone()[0]
        assert val == 2


def test_smoke_test_people_db_up_and_reset(postgres_people):
    """'people' db is up and got reset on every tables and sequences"""
    with BaseDb.connect(postgres_people.dsn).cursor() as cur:
        # tables are truncated after the first round
        cur.execute("select count(*) from dbversion")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 0

        # tables are truncated after the first round
        cur.execute("select count(*) from people")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 0

        # table and sequence are reset
        cur.execute("select count(*) from fun")
        nb_rows = cur.fetchone()[0]
        assert nb_rows == 0

        cur.execute("select nextval('serial')")
        val = cur.fetchone()[0]
        assert val == 1


# db with no initialization step, an empty db
postgres_no_init = postgresql_fact("postgresql_proc", dbname="something")


def test_smoke_test_db_no_init(postgres_no_init):
    """We can connect to the db nonetheless"""
    with BaseDb.connect(postgres_no_init.dsn).cursor() as cur:
        cur.execute("select now()")
        data = cur.fetchone()[0]
        assert data is not None



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import time
from typing import Dict, Iterator, List, Optional, Union

import pytest
import requests_mock

HTTP_GITHUB_API_URL = "https://api.github.com/repositories"


def fake_time_sleep(duration: float, sleep_calls: Optional[List[float]] = None):
    """Record calls to time.sleep in the sleep_calls list."""
    if duration < 0:
        raise ValueError("Can't sleep for a negative amount of time!")
    if sleep_calls is not None:
        sleep_calls.append(duration)


def fake_time_time():
    """Return 0 when running time.time()"""
    return 0


@pytest.fixture
def monkeypatch_sleep_calls(monkeypatch) -> Iterator[List[float]]:
    """Monkeypatch `time.time` and `time.sleep`. Returns a list cumulating the arguments
    passed to time.sleep()."""
    sleeps: List[float] = []
    monkeypatch.setattr(time, "sleep", lambda d: fake_time_sleep(d, sleeps))
    monkeypatch.setattr(time, "time", fake_time_time)
    yield sleeps


@pytest.fixture()
def num_before_ratelimit() -> int:
    """Number of successful requests before the ratelimit hits"""
    return 0


@pytest.fixture()
def num_ratelimit() -> Optional[int]:
    """Number of rate-limited requests; None means infinity"""
    return None


@pytest.fixture()
def ratelimit_reset() -> Optional[int]:
    """Value of the X-Ratelimit-Reset header on ratelimited responses"""
    return None


def github_ratelimit_callback(
    request: requests_mock.request._RequestObjectProxy,
    context: requests_mock.response._Context,
    ratelimit_reset: Optional[int],
) -> Dict[str, str]:
    """Return a rate-limited GitHub API response."""
    # Check request headers
    assert request.headers["Accept"] == "application/vnd.github.v3+json"
    assert request.headers["User-Agent"] is not None
    if "Authorization" in request.headers:
        context.status_code = 429
    else:
        context.status_code = 403

    if ratelimit_reset is not None:
        context.headers["X-Ratelimit-Reset"] = str(ratelimit_reset)

    return {
        "message": "API rate limit exceeded for <IP>.",
        "documentation_url": "https://developer.github.com/v3/#rate-limiting",
    }


def github_repo(i: int) -> Dict[str, Union[int, str]]:
    """Basic repository information returned by the GitHub API"""

    repo: Dict[str, Union[int, str]] = {
        "id": i,
        "html_url": f"https://github.com/origin/{i}",
    }

    # Set the pushed_at date on one of the origins
    if i == 4321:
        repo["pushed_at"] = "2018-11-08T13:16:24Z"

    return repo


def github_response_callback(
    request: requests_mock.request._RequestObjectProxy,
    context: requests_mock.response._Context,
    page_size: int = 1000,
    origin_count: int = 10000,
) -> List[Dict[str, Union[str, int]]]:
    """Return minimal GitHub API responses for the common case where the loader
    hasn't been rate-limited"""
    # Check request headers
    assert request.headers["Accept"] == "application/vnd.github.v3+json"
    assert request.headers["User-Agent"] is not None

    # Check request parameters: per_page == 1000, since = last_repo_id
    assert "per_page" in request.qs
    assert request.qs["per_page"] == [str(page_size)]
    assert "since" in request.qs

    since = int(request.qs["since"][0])

    next_page = since + page_size
    if next_page < origin_count:
        # the first id for the next page is within our origin count; add a Link
        # header to the response
        next_url = f"{HTTP_GITHUB_API_URL}?per_page={page_size}&since={next_page}"
        context.headers["Link"] = f"<{next_url}>; rel=next"

    return [github_repo(i) for i in range(since + 1, min(next_page, origin_count) + 1)]


@pytest.fixture()
def requests_ratelimited(
    num_before_ratelimit: int,
    num_ratelimit: Optional[int],
    ratelimit_reset: Optional[int],
) -> Iterator[requests_mock.Mocker]:
    """Mock requests to the GitHub API, returning a rate-limiting status code after
    `num_before_ratelimit` requests.

    GitHub does inconsistent rate-limiting:

    - Anonymous requests return a 403 status code
    - Authenticated requests return a 429 status code, with an X-Ratelimit-Reset header.

    This fixture takes multiple arguments (which can be overridden with a
    :func:`pytest.mark.parametrize` parameter):

    - num_before_ratelimit: the global number of requests until the ratelimit triggers
    - num_ratelimit: the number of requests that return a rate-limited response.
    - ratelimit_reset: the timestamp returned in X-Ratelimit-Reset if the request is
      authenticated.

    The default values set in the previous fixtures make all requests return a rate
    limit response.

    """
    current_request = 0

    def response_callback(request, context):
        nonlocal current_request
        current_request += 1
        if num_before_ratelimit < current_request and (
            num_ratelimit is None
            or current_request < num_before_ratelimit + num_ratelimit + 1
        ):
            return github_ratelimit_callback(request, context, ratelimit_reset)
        else:
            return github_response_callback(request, context)

    with requests_mock.Mocker() as mock:
        mock.get(HTTP_GITHUB_API_URL, json=response_callback)
        yield mock


@pytest.fixture
def github_credentials() -> List[Dict[str, str]]:
    """Return a static list of GitHub credentials"""
    return sorted(
        [{"username": f"swh{i:d}", "token": f"token-{i:d}"} for i in range(3)]
        + [
            {"username": f"swh-legacy{i:d}", "password": f"token-legacy-{i:d}"}
            for i in range(3)
        ],
        key=lambda c: c["username"],
    )


@pytest.fixture
def all_tokens(github_credentials) -> List[str]:
    """Return the list of tokens matching the static credential"""

    return [t.get("token", t.get("password")) for t in github_credentials]



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

import pytest
from sentry_sdk import capture_message

from swh.core.sentry import init_sentry, override_with_bool_envvar


@pytest.mark.parametrize(
    "envvalue,retval",
    (
        ("y", True),
        ("n", False),
        ("0", False),
        ("true", True),
        ("FaLsE", False),
        ("1", True),
    ),
)
def test_override_with_bool_envvar(monkeypatch, envvalue: str, retval: bool):
    """Test if the override_with_bool_envvar function returns appropriate results"""
    envvar = "OVERRIDE_WITH_BOOL_ENVVAR"
    monkeypatch.setenv(envvar, envvalue)
    for default in (True, False):
        assert override_with_bool_envvar(envvar, default) == retval


def test_override_with_bool_envvar_logging(monkeypatch, caplog):
    envvar = "OVERRIDE_WITH_BOOL_ENVVAR"
    monkeypatch.setenv(envvar, "not a boolean env value")
    for default in (True, False):
        caplog.clear()
        assert override_with_bool_envvar(envvar, default) == default
        assert len(caplog.records) == 1
        assert (
            "OVERRIDE_WITH_BOOL_ENVVAR='not a boolean env value'"
            in caplog.records[0].getMessage()
        )
        assert f"using default value {default}" in caplog.records[0].getMessage()
        assert caplog.records[0].levelname == "WARNING"


def test_sentry():
    reports = []
    init_sentry("http://example.org", extra_kwargs={"transport": reports.append})

    capture_message("Something went wrong")
    logging.error("Stupid error")

    assert len(reports) == 2
    assert reports[0]["message"] == "Something went wrong"
    assert reports[1]["logentry"]["message"] == "Stupid error"


def test_sentry_no_logging():
    reports = []
    init_sentry(
        "http://example.org",
        disable_logging_events=True,
        extra_kwargs={"transport": reports.append},
    )

    capture_message("Something went wrong")
    logging.error("Stupid error")

    assert len(reports) == 1
    assert reports[0]["message"] == "Something went wrong"


def test_sentry_no_logging_from_venv(monkeypatch):
    monkeypatch.setenv("SWH_SENTRY_DISABLE_LOGGING_EVENTS", "True")

    reports = []
    init_sentry(
        "http://example.org",
        extra_kwargs={"transport": reports.append},
    )

    capture_message("Something went wrong")
    logging.error("Stupid error")

    assert len(reports) == 1
    assert reports[0]["message"] == "Something went wrong"


def test_sentry_logging_from_venv(monkeypatch):
    monkeypatch.setenv("SWH_SENTRY_DISABLE_LOGGING_EVENTS", "false")

    reports = []
    init_sentry(
        "http://example.org",
        extra_kwargs={"transport": reports.append},
    )

    capture_message("Something went wrong")
    logging.error("Stupid error")

    assert len(reports) == 2



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import datetime, timedelta
from os import path

import pytest

from swh.core.cli.db import db as swhdb
from swh.core.db import BaseDb
from swh.core.db.db_utils import (
    get_database_info,
    get_sql_for_package,
    now,
    swh_db_module,
    swh_db_upgrade,
    swh_db_version,
    swh_db_versions,
    swh_set_db_module,
)

from .test_cli import craft_conninfo


@pytest.mark.parametrize("module", ["test.cli", "test.cli_new"])
def test_get_sql_for_package(mock_import_swhmodule, module):
    files = get_sql_for_package(module)
    assert files
    assert [f.name for f in files] == [
        "0-superuser-init.sql",
        "15-flavor.sql",
        "30-schema.sql",
        "40-funcs.sql",
        "50-data.sql",
    ]


@pytest.mark.parametrize("module", ["test.cli", "test.cli_new"])
def test_db_utils_versions(cli_runner, postgresql, mock_import_swhmodule, module):
    """Check get_database_info, swh_db_versions and swh_db_module work ok

    This test checks db versions for both a db with "new style" set of sql init
    scripts (i.e. the dbversion table is not created in these scripts, but by
    the populate_database_for_package() function directly, via the 'swh db
    init' command) and an "old style" set (dbversion created in the scripts)S.

    """
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(
        swhdb, ["init", module, "--dbname", conninfo, "--initial-version", 10]
    )
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    # check the swh_db_module() function
    assert swh_db_module(conninfo) == module

    # the dbversion and dbmodule tables exists and are populated
    dbmodule, dbversion, dbflavor = get_database_info(conninfo)
    # check also the swh_db_versions() function
    versions = swh_db_versions(conninfo)

    assert dbmodule == module
    assert dbversion == 10
    assert dbflavor == "default"
    # check also the swh_db_versions() function
    versions = swh_db_versions(conninfo)
    assert len(versions) == 1
    assert versions[0][0] == 10
    if module == "test.cli":
        assert versions[0][1] == datetime.fromisoformat(
            "2016-02-22T15:56:28.358587+00:00"
        )
        assert versions[0][2] == "Work In Progress"
    else:
        # new scheme but with no datastore (so no version support from there)
        assert versions[0][2] == "DB initialization"

    # add a few versions in dbversion
    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.executemany(
            "insert into dbversion(version, release, description) values (%s, %s, %s)",
            [(i, now(), f"Upgrade to version {i}") for i in range(11, 15)],
        )

    dbmodule, dbversion, dbflavor = get_database_info(conninfo)
    assert dbmodule == module
    assert dbversion == 14
    assert dbflavor == "default"

    versions = swh_db_versions(conninfo)
    assert len(versions) == 5
    for i, (version, ts, desc) in enumerate(versions):
        assert version == (14 - i)  # these are in reverse order
        if version > 10:
            assert desc == f"Upgrade to version {version}"
            assert (now() - ts) < timedelta(seconds=1)


@pytest.mark.parametrize("module", ["test.cli_new"])
def test_db_utils_upgrade(
    cli_runner, postgresql, mock_import_swhmodule, module, datadir
):
    """Check swh_db_upgrade"""
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    assert swh_db_version(conninfo) == 1
    new_version = swh_db_upgrade(conninfo, module)
    assert new_version == 6
    assert swh_db_version(conninfo) == 6

    versions = swh_db_versions(conninfo)
    # get rid of dates to ease checking
    versions = [(v[0], v[2]) for v in versions]
    assert versions[-1] == (1, "DB initialization")
    sqlbasedir = path.join(datadir, module.split(".", 1)[1], "sql", "upgrades")

    assert versions[1:-1] == [
        (i, f"Upgraded to version {i} using {sqlbasedir}/{i:03d}.sql")
        for i in range(5, 1, -1)
    ]
    assert versions[0] == (6, "Updated version from upgrade script")

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.execute("select url from origin where url like 'version%'")
        result = cur.fetchall()
        assert result == [("version%03d" % i,) for i in range(2, 7)]
        cur.execute(
            "select url from origin where url = 'this should never be executed'"
        )
        result = cur.fetchall()
        assert not result


@pytest.mark.parametrize("module", ["test.cli_new"])
def test_db_utils_swh_db_upgrade_sanity_checks(
    cli_runner, postgresql, mock_import_swhmodule, module, datadir
):
    """Check swh_db_upgrade"""
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.execute("drop table dbmodule")

    # try to upgrade with a unset module
    with pytest.raises(ValueError):
        swh_db_upgrade(conninfo, module)

    # check the dbmodule is unset
    assert swh_db_module(conninfo) is None

    # set the stored module to something else
    swh_set_db_module(conninfo, f"{module}2")
    assert swh_db_module(conninfo) == f"{module}2"

    # try to upgrade with a different module
    with pytest.raises(ValueError):
        swh_db_upgrade(conninfo, module)

    # revert to the proper module in the db
    swh_set_db_module(conninfo, module, force=True)
    assert swh_db_module(conninfo) == module
    # trying again is a noop
    swh_set_db_module(conninfo, module)
    assert swh_db_module(conninfo) == module

    # drop the dbversion table
    with cnx.transaction() as cur:
        cur.execute("drop table dbversion")
    # an upgrade should fail due to missing stored version
    with pytest.raises(ValueError):
        swh_db_upgrade(conninfo, module)


@pytest.mark.parametrize("module", ["test.cli", "test.cli_new"])
@pytest.mark.parametrize("flavor", [None, "default", "flavorA", "flavorB"])
def test_db_utils_flavor(cli_runner, postgresql, mock_import_swhmodule, module, flavor):
    """Check populate_database_for_package handle db flavor properly"""
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    cmd = ["init", module, "--dbname", conninfo]
    if flavor:
        cmd.extend(["--flavor", flavor])
    result = cli_runner.invoke(swhdb, cmd)
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    # check the swh_db_module() function
    assert swh_db_module(conninfo) == module

    # the dbversion and dbmodule tables exists and are populated
    dbmodule, _dbversion, dbflavor = get_database_info(conninfo)
    assert dbmodule == module
    assert dbflavor == (flavor or "default")



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from click.testing import CliRunner
from hypothesis import HealthCheck
import psycopg2
import pytest
from pytest_postgresql import factories

from swh.core.db.db_utils import import_swhmodule

os.environ["LC_ALL"] = "C.UTF-8"

# we use getattr here to keep mypy happy regardless hypothesis version
function_scoped_fixture_check = (
    [getattr(HealthCheck, "function_scoped_fixture")]
    if hasattr(HealthCheck, "function_scoped_fixture")
    else []
)


def create_role_guest(**kwargs):
    with psycopg2.connect(**kwargs) as conn:
        with conn.cursor() as cur:
            cur.execute("CREATE ROLE guest LOGIN PASSWORD 'guest'")


postgresql_proc = factories.postgresql_proc(
    load=[create_role_guest],
)


@pytest.fixture
def cli_runner():
    return CliRunner()


@pytest.fixture()
def mock_import_swhmodule(mocker, datadir):
    """This bypasses the module manipulation to make import_swhmodule return a mock
    object suitable for data test files listing via get_sql_for_package.

    For a given module `test.<mod>`, return a MagicMock object with a __name__
    set to `<mod>` and __file__ pointing to `data/<mod>/__init__.py`.

    The Mock object also defines a `get_datastore()` attribute on which the
    `current_version` attribute is set to 42.

    Typical usage::

      def test_xxx(cli_runner, mock_import_swhmodule):
        conninfo = craft_conninfo(test_db, "new-db")
        module_name = "test.cli"
        # the command below will use sql scripts from
        #     swh/core/db/tests/data/cli/sql/*.sql
        cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])

    """
    mock = mocker.MagicMock

    def import_swhmodule_mock(modname):
        if modname.startswith("test."):
            dirname = modname.split(".", 1)[1]

            def get_datastore(*args, **kw):
                return mock(current_version=42)

            return mock(
                __name__=modname,
                __file__=os.path.join(datadir, dirname, "__init__.py"),
                get_datastore=get_datastore,
            )
        else:
            return import_swhmodule(modname)

    return mocker.patch("swh.core.db.db_utils.import_swhmodule", import_swhmodule_mock)



# Copyright (C) 2015-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from contextlib import contextmanager
from datetime import datetime, timezone
import functools
from importlib import import_module
import logging
from os import path
import pathlib
import re
import subprocess
from typing import Collection, Dict, Iterator, List, Optional, Tuple, Union, cast

import psycopg2
import psycopg2.errors
import psycopg2.extensions
from psycopg2.extensions import connection as pgconnection
from psycopg2.extensions import encodings as pgencodings
from psycopg2.extensions import make_dsn
from psycopg2.extensions import parse_dsn as _parse_dsn

from swh.core.utils import numfile_sortkey as sortkey

logger = logging.getLogger(__name__)


def now():
    return datetime.now(tz=timezone.utc)


def stored_procedure(stored_proc):
    """decorator to execute remote stored procedure, specified as argument

    Generally, the body of the decorated function should be empty. If it is
    not, the stored procedure will be executed first; the function body then.

    """

    def wrap(meth):
        @functools.wraps(meth)
        def _meth(self, *args, **kwargs):
            cur = kwargs.get("cur", None)
            self._cursor(cur).execute("SELECT %s()" % stored_proc)
            meth(self, *args, **kwargs)

        return _meth

    return wrap


def jsonize(value):
    """Convert a value to a psycopg2 JSON object if necessary"""
    if isinstance(value, dict):
        return psycopg2.extras.Json(value)

    return value


@contextmanager
def connect_to_conninfo(
    db_or_conninfo: Union[str, pgconnection]
) -> Iterator[pgconnection]:
    """Connect to the database passed as argument.

    Args:
        db_or_conninfo: A database connection, or a database connection info string

    Returns:
        a connected database handle or None if the database is not initialized

    """
    if isinstance(db_or_conninfo, pgconnection):
        yield db_or_conninfo
    else:
        if "=" not in db_or_conninfo and "//" not in db_or_conninfo:
            # Database name
            db_or_conninfo = f"dbname={db_or_conninfo}"

    try:
        db = psycopg2.connect(db_or_conninfo)
    except psycopg2.Error:
        logger.exception("Failed to connect to `%s`", db_or_conninfo)
    else:
        yield db


def swh_db_version(db_or_conninfo: Union[str, pgconnection]) -> Optional[int]:
    """Retrieve the swh version of the database.

    If the database is not initialized, this logs a warning and returns None.

    Args:
      db_or_conninfo: A database connection, or a database connection info string

    Returns:
        Either the version of the database, or None if it couldn't be detected
    """
    try:
        with connect_to_conninfo(db_or_conninfo) as db:
            if not db:
                return None
            with db.cursor() as c:
                query = "select version from dbversion order by dbversion desc limit 1"
                try:
                    c.execute(query)
                    result = c.fetchone()
                    if result:
                        return result[0]
                except psycopg2.errors.UndefinedTable:
                    return None
    except Exception:
        logger.exception("Could not get version from `%s`", db_or_conninfo)
    return None


def swh_db_versions(
    db_or_conninfo: Union[str, pgconnection]
) -> Optional[List[Tuple[int, datetime, str]]]:
    """Retrieve the swh version history of the database.

    If the database is not initialized, this logs a warning and returns None.

    Args:
      db_or_conninfo: A database connection, or a database connection info string

    Returns:
        Either the version of the database, or None if it couldn't be detected
    """
    try:
        with connect_to_conninfo(db_or_conninfo) as db:
            if not db:
                return None
            with db.cursor() as c:
                query = (
                    "select version, release, description "
                    "from dbversion order by dbversion desc"
                )
                try:
                    c.execute(query)
                    return cast(List[Tuple[int, datetime, str]], c.fetchall())
                except psycopg2.errors.UndefinedTable:
                    return None
    except Exception:
        logger.exception("Could not get versions from `%s`", db_or_conninfo)
        return None


def swh_db_upgrade(
    conninfo: str, modname: str, to_version: Optional[int] = None
) -> int:
    """Upgrade the database at `conninfo` for module `modname`

    This will run migration scripts found in the `sql/upgrades` subdirectory of
    the module `modname`. By default, this will upgrade to the latest declared version.

    Args:
      conninfo: A database connection, or a database connection info string
      modname: datastore module the database stores content for
      to_version: if given, update the database to this version rather than the latest

    """

    if to_version is None:
        to_version = 99999999

    db_module, db_version, db_flavor = get_database_info(conninfo)
    if db_version is None:
        raise ValueError("Unable to retrieve the current version of the database")
    if db_module is None:
        raise ValueError("Unable to retrieve the module of the database")
    if db_module != modname:
        raise ValueError(
            "The stored module of the database is different than the given one"
        )

    sqlfiles = [
        fname
        for fname in get_sql_for_package(modname, upgrade=True)
        if db_version < int(fname.stem) <= to_version
    ]

    if not sqlfiles:
        return db_version

    for sqlfile in sqlfiles:
        new_version = int(path.splitext(path.basename(sqlfile))[0])
        logger.info("Executing migration script '%s'", sqlfile)
        if db_version is not None and (new_version - db_version) > 1:
            logger.error(
                f"There are missing migration steps between {db_version} and "
                f"{new_version}. It might be expected but it most unlikely is not. "
                "Will stop here."
            )
            return db_version

        execute_sqlfiles([sqlfile], conninfo, db_flavor)

        # check if the db version has been updated by the upgrade script
        db_version = swh_db_version(conninfo)
        assert db_version is not None
        if db_version == new_version:
            # nothing to do, upgrade script did the job
            pass
        elif db_version == new_version - 1:
            # it has not (new style), so do it
            swh_set_db_version(
                conninfo,
                new_version,
                desc=f"Upgraded to version {new_version} using {sqlfile}",
            )
            db_version = swh_db_version(conninfo)
        else:
            # upgrade script did it wrong
            logger.error(
                f"The upgrade script {sqlfile} did not update the dbversion table "
                f"consistently ({db_version} vs. expected {new_version}). "
                "Will stop migration here. Please check your migration scripts."
            )
            return db_version
    return new_version


def swh_db_module(db_or_conninfo: Union[str, pgconnection]) -> Optional[str]:
    """Retrieve the swh module used to create the database.

    If the database is not initialized, this logs a warning and returns None.

    Args:
      db_or_conninfo: A database connection, or a database connection info string

    Returns:
        Either the module of the database, or None if it couldn't be detected
    """
    try:
        with connect_to_conninfo(db_or_conninfo) as db:
            if not db:
                return None
            with db.cursor() as c:
                query = "select dbmodule from dbmodule limit 1"
                try:
                    c.execute(query)
                    resp = c.fetchone()
                    if resp:
                        return resp[0]
                except psycopg2.errors.UndefinedTable:
                    return None
    except Exception:
        logger.exception("Could not get module from `%s`", db_or_conninfo)
    return None


def swh_set_db_module(
    db_or_conninfo: Union[str, pgconnection], module: str, force=False
) -> None:
    """Set the swh module used to create the database.

    Fails if the dbmodule is already set or the table does not exist.

    Args:
      db_or_conninfo: A database connection, or a database connection info string
      module: the swh module to register (without the leading 'swh.')
    """
    update = False
    if module.startswith("swh."):
        module = module[4:]

    current_module = swh_db_module(db_or_conninfo)
    if current_module is not None:
        if current_module == module:
            logger.warning("The database module is already set to %s", module)
            return

        if not force:
            raise ValueError(
                "The database module is already set to a value %s "
                "different than given %s",
                current_module,
                module,
            )
        # force is True
        update = True

    with connect_to_conninfo(db_or_conninfo) as db:
        if not db:
            return None

        sqlfiles = [
            fname
            for fname in get_sql_for_package("swh.core.db")
            if "dbmodule" in fname.stem
        ]
        execute_sqlfiles(sqlfiles, db_or_conninfo)

        with db.cursor() as c:
            if update:
                query = "update dbmodule set dbmodule = %s"
            else:
                query = "insert into dbmodule(dbmodule) values (%s)"
            c.execute(query, (module,))
        db.commit()


def swh_set_db_version(
    db_or_conninfo: Union[str, pgconnection],
    version: int,
    ts: Optional[datetime] = None,
    desc: str = "Work in progress",
) -> None:
    """Set the version of the database.

    Fails if the dbversion table does not exists.

    Args:
      db_or_conninfo: A database connection, or a database connection info string
      version: the version to add
    """
    if ts is None:
        ts = now()

    with connect_to_conninfo(db_or_conninfo) as db:
        if not db:
            return None
        with db.cursor() as c:
            query = (
                "insert into dbversion(version, release, description) "
                "values (%s, %s, %s)"
            )
            c.execute(query, (version, ts, desc))
            db.commit()


def swh_db_flavor(db_or_conninfo: Union[str, pgconnection]) -> Optional[str]:
    """Retrieve the swh flavor of the database.

    If the database is not initialized, or the database doesn't support
    flavors, this returns None.

    Args:
      db_or_conninfo: A database connection, or a database connection info string

    Returns:
        The flavor of the database, or None if it could not be detected.
    """
    try:
        with connect_to_conninfo(db_or_conninfo) as db:
            if not db:
                return None
            with db.cursor() as c:
                query = "select swh_get_dbflavor()"
                try:
                    c.execute(query)
                    result = c.fetchone()
                    assert result is not None  # to keep mypy happy
                    return result[0]
                except psycopg2.errors.UndefinedFunction:
                    # function not found: no flavor
                    return None
    except Exception:
        logger.exception("Could not get flavor from `%s`", db_or_conninfo)
        return None


# The following code has been imported from psycopg2, version 2.7.4,
# https://github.com/psycopg/psycopg2/tree/5afb2ce803debea9533e293eef73c92ffce95bcd
# and modified by Software Heritage.
#
# Original file: lib/extras.py
#
# psycopg2 is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 3 of the License, or (at your option) any
# later version.


def _paginate(seq, page_size):
    """Consume an iterable and return it in chunks.
    Every chunk is at most `page_size`. Never return an empty chunk.
    """
    page = []
    it = iter(seq)
    while 1:
        try:
            for i in range(page_size):
                page.append(next(it))
            yield page
            page = []
        except StopIteration:
            if page:
                yield page
            return


def _split_sql(sql):
    """Split *sql* on a single ``%s`` placeholder.
    Split on the %s, perform %% replacement and return pre, post lists of
    snippets.
    """
    curr = pre = []
    post = []
    tokens = re.split(rb"(%.)", sql)
    for token in tokens:
        if len(token) != 2 or token[:1] != b"%":
            curr.append(token)
            continue

        if token[1:] == b"s":
            if curr is pre:
                curr = post
            else:
                raise ValueError("the query contains more than one '%s' placeholder")
        elif token[1:] == b"%":
            curr.append(b"%")
        else:
            raise ValueError(
                "unsupported format character: '%s'"
                % token[1:].decode("ascii", "replace")
            )

    if curr is pre:
        raise ValueError("the query doesn't contain any '%s' placeholder")

    return pre, post


def execute_values_generator(cur, sql, argslist, template=None, page_size=100):
    """Execute a statement using SQL ``VALUES`` with a sequence of parameters.
    Rows returned by the query are returned through a generator.
    You need to consume the generator for the queries to be executed!

    :param cur: the cursor to use to execute the query.
    :param sql: the query to execute. It must contain a single ``%s``
        placeholder, which will be replaced by a `VALUES list`__.
        Example: ``"INSERT INTO mytable (id, f1, f2) VALUES %s"``.
    :param argslist: sequence of sequences or dictionaries with the arguments
        to send to the query. The type and content must be consistent with
        *template*.
    :param template: the snippet to merge to every item in *argslist* to
        compose the query.

        - If the *argslist* items are sequences it should contain positional
          placeholders (e.g. ``"(%s, %s, %s)"``, or ``"(%s, %s, 42)``" if there
          are constants value...).
        - If the *argslist* items are mappings it should contain named
          placeholders (e.g. ``"(%(id)s, %(f1)s, 42)"``).

        If not specified, assume the arguments are sequence and use a simple
        positional template (i.e.  ``(%s, %s, ...)``), with the number of
        placeholders sniffed by the first element in *argslist*.
    :param page_size: maximum number of *argslist* items to include in every
        statement. If there are more items the function will execute more than
        one statement.
    :param yield_from_cur: Whether to yield results from the cursor in this
        function directly.

    .. __: https://www.postgresql.org/docs/current/static/queries-values.html

    After the execution of the function the `cursor.rowcount` property will
    **not** contain a total result.
    """
    # we can't just use sql % vals because vals is bytes: if sql is bytes
    # there will be some decoding error because of stupid codec used, and Py3
    # doesn't implement % on bytes.
    if not isinstance(sql, bytes):
        sql = sql.encode(pgencodings[cur.connection.encoding])
    pre, post = _split_sql(sql)

    for page in _paginate(argslist, page_size=page_size):
        if template is None:
            template = b"(" + b",".join([b"%s"] * len(page[0])) + b")"
        parts = pre[:]
        for args in page:
            parts.append(cur.mogrify(template, args))
            parts.append(b",")
        parts[-1:] = post
        cur.execute(b"".join(parts))
        yield from cur


def import_swhmodule(modname):
    if not modname.startswith("swh."):
        modname = f"swh.{modname}"
    try:
        m = import_module(modname)
    except ImportError as exc:
        logger.error(f"Could not load the {modname} module: {exc}")
        return None
    return m


def get_sql_for_package(modname: str, upgrade: bool = False) -> List[pathlib.Path]:
    """Return the (sorted) list of sql script files for the given swh module

    If upgrade is True, return the list of available migration scripts,
    otherwise, return the list of initialization scripts.
    """
    m = import_swhmodule(modname)
    if m is None:
        raise ValueError(f"Module {modname} cannot be loaded")

    sqldir = pathlib.Path(m.__file__).parent / "sql"
    if upgrade:
        sqldir /= "upgrades"
    if not sqldir.is_dir():
        raise ValueError(
            "Module {} does not provide a db schema (no sql/ dir)".format(modname)
        )
    return sorted(sqldir.glob("*.sql"), key=lambda x: sortkey(x.name))


def populate_database_for_package(
    modname: str, conninfo: str, flavor: Optional[str] = None
) -> Tuple[bool, Optional[int], Optional[str]]:
    """Populate the database, pointed at with ``conninfo``,
    using the SQL files found in the package ``modname``.
    Also fill the 'dbmodule' table with the given ``modname``.

    Args:
      modname: Name of the module of which we're loading the files
      conninfo: connection info string for the SQL database
      flavor: the module-specific flavor which we want to initialize the database under

    Returns:
      Tuple with three elements: whether the database has been initialized; the current
      version of the database; if it exists, the flavor of the database.
    """
    current_version = swh_db_version(conninfo)
    if current_version is not None:
        dbflavor = swh_db_flavor(conninfo)
        return False, current_version, dbflavor

    def globalsortkey(key):
        "like sortkey but only on basenames"
        return sortkey(path.basename(key))

    sqlfiles = get_sql_for_package(modname) + get_sql_for_package("swh.core.db")
    sqlfiles = sorted(sqlfiles, key=lambda x: sortkey(x.stem))
    sqlfiles = [fpath for fpath in sqlfiles if "-superuser-" not in fpath.stem]
    execute_sqlfiles(sqlfiles, conninfo, flavor)

    # populate the dbmodule table
    swh_set_db_module(conninfo, modname)

    current_db_version = swh_db_version(conninfo)
    dbflavor = swh_db_flavor(conninfo)
    return True, current_db_version, dbflavor


def initialize_database_for_module(
    modname: str, version: int, flavor: Optional[str] = None, **kwargs
):
    """Helper function to initialize and populate a database for the given module

    This aims at helping the usage of pytest_postgresql for swh.core.db based datastores.
    Typical usage will be (here for swh.storage)::

      from pytest_postgresql import factories

      storage_postgresql_proc = factories.postgresql_proc(
        load=[partial(initialize_database_for_module, modname="storage", version=42)]
      )
      storage_postgresql = factories.postgresql("storage_postgresql_proc")

    """
    conninfo = psycopg2.connect(**kwargs).dsn
    init_admin_extensions(modname, conninfo)
    populate_database_for_package(modname, conninfo, flavor)
    try:
        swh_set_db_version(conninfo, version)
    except psycopg2.errors.UniqueViolation:
        logger.warn(
            "Version already set by db init scripts. "
            f"This generally means the swh.{modname} package needs to be "
            "updated for swh.core>=1.2"
        )


def get_database_info(
    conninfo: str,
) -> Tuple[Optional[str], Optional[int], Optional[str]]:
    """Get version, flavor and module of the db"""
    dbmodule = swh_db_module(conninfo)
    dbversion = swh_db_version(conninfo)
    dbflavor = None
    if dbversion is not None:
        dbflavor = swh_db_flavor(conninfo)
    return (dbmodule, dbversion, dbflavor)


def parse_dsn_or_dbname(dsn_or_dbname: str) -> Dict[str, str]:
    """Parse a psycopg2 dsn, falling back to supporting plain database names as well"""
    try:
        return _parse_dsn(dsn_or_dbname)
    except psycopg2.ProgrammingError:
        # psycopg2 failed to parse the DSN; it's probably a database name,
        # handle it as such
        return _parse_dsn(f"dbname={dsn_or_dbname}")


def init_admin_extensions(modname: str, conninfo: str) -> None:
    """The remaining initialization process -- running -superuser- SQL files -- is done
    using the given conninfo, thus connecting to the newly created database

    """
    sqlfiles = get_sql_for_package(modname)
    sqlfiles = [fname for fname in sqlfiles if "-superuser-" in fname.stem]
    execute_sqlfiles(sqlfiles, conninfo)


def create_database_for_package(
    modname: str, conninfo: str, template: str = "template1"
):
    """Create the database pointed at with ``conninfo``, and initialize it using
    -superuser- SQL files found in the package ``modname``.

    Args:
      modname: Name of the module of which we're loading the files
      conninfo: connection info string or plain database name for the SQL database
      template: the name of the database to connect to and use as template to create
        the new database

    """
    # Use the given conninfo string, but with dbname replaced by the template dbname
    # for the database creation step
    creation_dsn = parse_dsn_or_dbname(conninfo)
    dbname = creation_dsn["dbname"]
    creation_dsn["dbname"] = template
    logger.debug("db_create dbname=%s (from %s)", dbname, template)
    subprocess.check_call(
        [
            "psql",
            "--quiet",
            "--no-psqlrc",
            "-v",
            "ON_ERROR_STOP=1",
            "-d",
            make_dsn(**creation_dsn),
            "-c",
            f'CREATE DATABASE "{dbname}"',
        ]
    )
    init_admin_extensions(modname, conninfo)


def execute_sqlfiles(
    sqlfiles: Collection[pathlib.Path],
    db_or_conninfo: Union[str, pgconnection],
    flavor: Optional[str] = None,
):
    """Execute a list of SQL files on the database pointed at with ``db_or_conninfo``.

    Args:
      sqlfiles: List of SQL files to execute
      db_or_conninfo: A database connection, or a database connection info string
      flavor: the database flavor to initialize
    """
    if isinstance(db_or_conninfo, str):
        conninfo = db_or_conninfo
    else:
        conninfo = db_or_conninfo.dsn

    psql_command = [
        "psql",
        "--quiet",
        "--no-psqlrc",
        "-v",
        "ON_ERROR_STOP=1",
        "-d",
        conninfo,
    ]

    flavor_set = False
    for sqlfile in sqlfiles:
        logger.debug(f"execute SQL file {sqlfile} dbname={conninfo}")
        subprocess.check_call(psql_command + ["-f", str(sqlfile)])

        if (
            flavor is not None
            and not flavor_set
            and sqlfile.name.endswith("-flavor.sql")
        ):
            logger.debug("Setting database flavor %s", flavor)
            query = f"insert into dbflavor (flavor) values ('{flavor}')"
            subprocess.check_call(psql_command + ["-c", query])
            flavor_set = True

    if flavor is not None and not flavor_set:
        logger.warn(
            "Asked for flavor %s, but module does not support database flavors",
            flavor,
        )

    # Grant read-access to guest user on all tables of the schema (if possible)
    with connect_to_conninfo(db_or_conninfo) as db:
        try:
            with db.cursor() as c:
                query = "grant select on all tables in schema public to guest"
                c.execute(query)
        except Exception:
            logger.warning("Grant read-only access to guest user failed. Skipping.")



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import os
import traceback

import pytest
import yaml

from swh.core.cli.db import db as swhdb
from swh.core.db import BaseDb
from swh.core.db.db_utils import import_swhmodule, swh_db_module, swh_db_version
from swh.core.tests.test_cli import assert_section_contains


def test_cli_swh_help(swhmain, cli_runner):
    swhmain.add_command(swhdb)
    result = cli_runner.invoke(swhmain, ["-h"])
    assert result.exit_code == 0
    assert_section_contains(
        result.output, "Commands", "db  Software Heritage database generic tools."
    )


help_db_snippets = (
    (
        "Usage",
        (
            "Usage: swh db [OPTIONS] COMMAND [ARGS]...",
            "Software Heritage database generic tools.",
        ),
    ),
    (
        "Commands",
        (
            "create      Create a database for the Software Heritage <module>.",
            "init        Initialize a database for the Software Heritage <module>.",
            "init-admin  Execute superuser-level initialization steps",
        ),
    ),
)


def test_cli_swh_db_help(swhmain, cli_runner):
    swhmain.add_command(swhdb)
    result = cli_runner.invoke(swhmain, ["db", "-h"])
    assert result.exit_code == 0
    for section, snippets in help_db_snippets:
        for snippet in snippets:
            assert_section_contains(result.output, section, snippet)


@pytest.fixture
def swh_db_cli(cli_runner, monkeypatch, postgresql):
    """This initializes a cli_runner and sets the correct environment variable expected by
    the cli to run appropriately (when not specifying the --dbname flag)

    """
    db_params = postgresql.get_dsn_parameters()
    monkeypatch.setenv("PGHOST", db_params["host"])
    monkeypatch.setenv("PGUSER", db_params["user"])
    monkeypatch.setenv("PGPORT", db_params["port"])

    return cli_runner, db_params


def craft_conninfo(test_db, dbname=None) -> str:
    """Craft conninfo string out of the test_db object. This also allows to override the
    dbname."""
    db_params = test_db.get_dsn_parameters()
    if dbname:
        params = copy.deepcopy(db_params)
        params["dbname"] = dbname
    else:
        params = db_params
    return "postgresql://{user}@{host}:{port}/{dbname}".format(**params)


def test_cli_swh_db_create_and_init_db(cli_runner, postgresql, mock_import_swhmodule):
    """Create a db then initializing it should be ok"""
    module_name = "test.cli"

    conninfo = craft_conninfo(postgresql, "new-db")
    # This creates the db and installs the necessary admin extensions
    result = cli_runner.invoke(swhdb, ["create", module_name, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    # This initializes the schema and data
    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])

    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    # the origin value in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, installed during db creation step)
    with BaseDb.connect(conninfo).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_initialization_fail_without_creation_first(
    cli_runner, postgresql, mock_import_swhmodule
):
    """Init command on an inexisting db cannot work"""
    module_name = "test.cli"  # it's mocked here
    conninfo = craft_conninfo(postgresql, "inexisting-db")

    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    # Fails because we cannot connect to an inexisting db
    assert result.exit_code == 1, f"Unexpected output: {result.output}"


def test_cli_swh_db_initialization_fail_without_extension(
    cli_runner, postgresql, mock_import_swhmodule
):
    """Init command cannot work without privileged extension.

    In this test, the schema needs privileged extension to work.

    """
    module_name = "test.cli"  # it's mocked here
    conninfo = craft_conninfo(postgresql)

    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])
    # Fails as the function `public.digest` is not installed, init-admin calls is needed
    # first (the next tests show such behavior)
    assert result.exit_code == 1, f"Unexpected output: {result.output}"


def test_cli_swh_db_initialization_works_with_flags(
    cli_runner, postgresql, mock_import_swhmodule
):
    """Init commands with carefully crafted libpq conninfo works"""
    module_name = "test.cli"  # it's mocked here
    conninfo = craft_conninfo(postgresql)

    result = cli_runner.invoke(swhdb, ["init-admin", module_name, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])

    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    # the origin values in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, init-admin calls installs it)
    with BaseDb.connect(postgresql.dsn).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_initialization_with_env(
    swh_db_cli, mock_import_swhmodule, postgresql
):
    """Init commands with standard environment variables works"""
    module_name = "test.cli"  # it's mocked here
    cli_runner, db_params = swh_db_cli
    result = cli_runner.invoke(
        swhdb, ["init-admin", module_name, "--dbname", db_params["dbname"]]
    )
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    result = cli_runner.invoke(
        swhdb, ["init", module_name, "--dbname", db_params["dbname"]]
    )

    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    # the origin values in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, init-admin calls installs it)
    with BaseDb.connect(postgresql.dsn).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_initialization_idempotent(
    swh_db_cli, mock_import_swhmodule, postgresql
):
    """Multiple runs of the init commands are idempotent"""
    module_name = "test.cli"  # mocked
    cli_runner, db_params = swh_db_cli

    result = cli_runner.invoke(
        swhdb, ["init-admin", module_name, "--dbname", db_params["dbname"]]
    )
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    result = cli_runner.invoke(
        swhdb, ["init", module_name, "--dbname", db_params["dbname"]]
    )
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    result = cli_runner.invoke(
        swhdb, ["init-admin", module_name, "--dbname", db_params["dbname"]]
    )
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    result = cli_runner.invoke(
        swhdb, ["init", module_name, "--dbname", db_params["dbname"]]
    )
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    # the origin values in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, init-admin calls installs it)
    with BaseDb.connect(postgresql.dsn).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_create_and_init_db_new_api(
    cli_runner, postgresql, mock_import_swhmodule, mocker, tmp_path
):
    """Create a db then initializing it should be ok for a "new style" datastore"""
    module_name = "test.cli_new"

    conninfo = craft_conninfo(postgresql)

    # This initializes the schema and data
    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(yaml.dump({module_name: {"cls": "postgresql", "db": conninfo}}))
    result = cli_runner.invoke(swhdb, ["init-admin", module_name, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init", module_name])

    assert (
        result.exit_code == 0
    ), f"Unexpected output: {traceback.print_tb(result.exc_info[2])}"

    # the origin value in the scripts uses a hash function (which implementation wise
    # uses a function from the pgcrypt extension, installed during db creation step)
    with BaseDb.connect(conninfo).cursor() as cur:
        cur.execute("select * from origin")
        origins = cur.fetchall()
        assert len(origins) == 1


def test_cli_swh_db_upgrade_new_api(cli_runner, postgresql, datadir, mocker, tmp_path):
    """Upgrade scenario for a "new style" datastore"""
    module_name = "test.cli_new"

    # the `current_version` variable is the version that will be returned by
    # any call to `get_current_version()` in this test session, thanks to the
    # local mocked version of import_swhmodule() below.
    current_version = 1

    # custom version of the mockup to make it easy to change the
    # current_version returned by get_current_version()
    # TODO: find a better solution for this...
    def import_swhmodule_mock(modname):
        if modname.startswith("test."):
            dirname = modname.split(".", 1)[1]

            def get_datastore(cls, **kw):
                return mocker.MagicMock(current_version=current_version)

            return mocker.MagicMock(
                __name__=modname,
                __file__=os.path.join(datadir, dirname, "__init__.py"),
                name=modname,
                get_datastore=get_datastore,
            )

        return import_swhmodule(modname)

    mocker.patch("swh.core.db.db_utils.import_swhmodule", import_swhmodule_mock)
    conninfo = craft_conninfo(postgresql)

    # This initializes the schema and data
    cfgfile = tmp_path / "config.yml"
    cfgfile.write_text(yaml.dump({module_name: {"cls": "postgresql", "db": conninfo}}))
    result = cli_runner.invoke(swhdb, ["init-admin", module_name, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "init", module_name])

    assert (
        result.exit_code == 0
    ), f"Unexpected output: {traceback.print_tb(result.exc_info[2])}"

    assert swh_db_version(conninfo) == 1

    # the upgrade should not do anything because the datastore does advertise
    # version 1
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", module_name])
    assert swh_db_version(conninfo) == 1

    # advertise current version as 3, a simple upgrade should get us there, but
    # no further
    current_version = 3
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", module_name])
    assert swh_db_version(conninfo) == 3

    # an attempt to go further should not do anything
    result = cli_runner.invoke(
        swhdb, ["-C", cfgfile, "upgrade", module_name, "--to-version", 5]
    )
    assert swh_db_version(conninfo) == 3
    # an attempt to go lower should not do anything
    result = cli_runner.invoke(
        swhdb, ["-C", cfgfile, "upgrade", module_name, "--to-version", 2]
    )
    assert swh_db_version(conninfo) == 3

    # advertise current version as 6, an upgrade with --to-version 4 should
    # stick to the given version 4 and no further
    current_version = 6
    result = cli_runner.invoke(
        swhdb, ["-C", cfgfile, "upgrade", module_name, "--to-version", 4]
    )
    assert swh_db_version(conninfo) == 4
    assert "migration was not complete" in result.output

    # attempt to upgrade to a newer version than current code version fails
    result = cli_runner.invoke(
        swhdb,
        ["-C", cfgfile, "upgrade", module_name, "--to-version", current_version + 1],
    )
    assert result.exit_code != 0
    assert swh_db_version(conninfo) == 4

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.execute("drop table dbmodule")
    assert swh_db_module(conninfo) is None

    # db migration should recreate the missing dbmodule table
    result = cli_runner.invoke(swhdb, ["-C", cfgfile, "upgrade", module_name])
    assert result.exit_code == 0
    assert "Warning: the database does not have a dbmodule table." in result.output
    assert (
        "Write the module information (test.cli_new) in the database? [Y/n]"
        in result.output
    )
    assert swh_db_module(conninfo) == module_name


def test_cli_swh_db_version(swh_db_cli, mock_import_swhmodule, postgresql):
    module_name = "test.cli"
    cli_runner, db_params = swh_db_cli

    conninfo = craft_conninfo(postgresql, "test-db-version")
    # This creates the db and installs the necessary admin extensions
    result = cli_runner.invoke(swhdb, ["create", module_name, "--dbname", conninfo])
    assert result.exit_code == 0, f"Unexpected output: {result.output}"

    # This initializes the schema and data
    result = cli_runner.invoke(swhdb, ["init", module_name, "--dbname", conninfo])

    actual_db_version = swh_db_version(conninfo)

    with BaseDb.connect(conninfo).cursor() as cur:
        cur.execute("select version from dbversion order by version desc limit 1")
        expected_version = cur.fetchone()[0]
        assert actual_db_version == expected_version

    assert result.exit_code == 0, f"Unexpected output: {result.output}"
    assert (
        f"initialized (flavor default) at version {expected_version}" in result.output
    )



#!/usr/bin/env python3
# Copyright (C) 2018-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from os import environ
import warnings

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

warnings.filterwarnings("ignore")  # noqa prevent psycopg from side-tracking us


logger = logging.getLogger(__name__)


@swh_cli_group.group(name="db", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(exists=True, dir_okay=False),
    help="Configuration file.",
)
@click.pass_context
def db(ctx, config_file):
    """Software Heritage database generic tools."""
    from swh.core.config import read as config_read

    ctx.ensure_object(dict)
    if config_file is None:
        config_file = environ.get("SWH_CONFIG_FILENAME")
    cfg = config_read(config_file)
    ctx.obj["config"] = cfg


@db.command(name="create", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name.",
    default="softwareheritage-dev",
    show_default=True,
)
@click.option(
    "--template",
    "-T",
    help="Template database from which to build this database.",
    default="template1",
    show_default=True,
)
def db_create(module, dbname, template):
    """Create a database for the Software Heritage <module>.

    and potentially execute superuser-level initialization steps.

    Example::

        swh db create -d swh-test storage

    If you want to specify non-default postgresql connection parameters, please
    provide them using standard environment variables or by the mean of a
    properly crafted libpq connection URI. See psql(1) man page (section
    ENVIRONMENTS) for details.

    Note: this command requires a postgresql connection with superuser permissions.

    Example::

        PGPORT=5434 swh db create indexer
        swh db create -d postgresql://superuser:passwd@pghost:5433/swh-storage storage

    """
    from swh.core.db.db_utils import create_database_for_package

    logger.debug("db_create %s dn_name=%s", module, dbname)
    create_database_for_package(module, dbname, template)


@db.command(name="init-admin", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name.",
    default="softwareheritage-dev",
    show_default=True,
)
def db_init_admin(module: str, dbname: str) -> None:
    """Execute superuser-level initialization steps (e.g pg extensions, admin functions,
    ...)

    Example::

        PGPASSWORD=... swh db init-admin -d swh-test scheduler

    If you want to specify non-default postgresql connection parameters, please
    provide them using standard environment variables or by the mean of a
    properly crafted libpq connection URI. See psql(1) man page (section
    ENVIRONMENTS) for details.

    Note: this command requires a postgresql connection with superuser permissions (e.g
    postgres, swh-admin, ...)

    Example::

        PGPORT=5434 swh db init-admin scheduler
        swh db init-admin -d postgresql://superuser:passwd@pghost:5433/swh-scheduler \
          scheduler

    """
    from swh.core.db.db_utils import init_admin_extensions

    logger.debug("db_init_admin %s dbname=%s", module, dbname)
    init_admin_extensions(module, dbname)


@db.command(name="init", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--dbname",
    "--db-name",
    "-d",
    help="Database name or connection URI.",
    default=None,
    show_default=False,
)
@click.option(
    "--flavor",
    help="Database flavor.",
    default=None,
)
@click.option(
    "--initial-version", help="Database initial version.", default=1, show_default=True
)
@click.pass_context
def db_init(ctx, module, dbname, flavor, initial_version):
    """Initialize a database for the Software Heritage <module>.

    The database connection string comes from the configuration file (see
    option ``--config-file`` in ``swh db --help``) in the section named after
    the MODULE argument.

    Example::

        $ cat conf.yml
        storage:
          cls: postgresql
          db: postgresql://user:passwd@pghost:5433/swh-storage
          objstorage:
            cls: memory

        $ swh db -C conf.yml init storage  # or
        $ SWH_CONFIG_FILENAME=conf.yml swh db init storage

    Note that the connection string can also be passed directly using the
    '--db-name' option, but this usage is about to be deprecated.

    """
    from swh.core.db.db_utils import (
        get_database_info,
        import_swhmodule,
        populate_database_for_package,
        swh_set_db_version,
    )

    cfg = None
    if dbname is None:
        # use the db cnx from the config file; the expected config entry is the
        # given module name
        cfg = ctx.obj["config"].get(module, {})
        dbname = get_dburl_from_config(cfg)

    if not dbname:
        raise click.BadParameter(
            "Missing the postgresql connection configuration. Either fix your "
            "configuration file or use the --dbname option."
        )

    logger.debug("db_init %s flavor=%s dbname=%s", module, flavor, dbname)

    initialized, dbversion, dbflavor = populate_database_for_package(
        module, dbname, flavor
    )
    if dbversion is None:
        if cfg is not None:
            # db version has not been populated by sql init scripts (new style),
            # let's do it; instantiate the data source to retrieve the current
            # (expected) db version
            datastore_factory = getattr(import_swhmodule(module), "get_datastore", None)
            if datastore_factory:
                datastore = datastore_factory(**cfg)
                if not hasattr(datastore, "current_version"):
                    logger.warning(
                        "Datastore %s does not declare the "
                        "'current_version' attribute",
                        datastore,
                    )
                else:
                    code_version = datastore.current_version
                    logger.info(
                        "Initializing database version to %s from the %s datastore",
                        code_version,
                        module,
                    )
                    swh_set_db_version(dbname, code_version, desc="DB initialization")

    dbversion = get_database_info(dbname)[1]
    if dbversion is None:
        logger.info(
            "Initializing database version to %s "
            "from the command line option --initial-version",
            initial_version,
        )
        swh_set_db_version(dbname, initial_version, desc="DB initialization")

    dbversion = get_database_info(dbname)[1]
    assert dbversion is not None

    # TODO: Ideally migrate the version from db_version to the latest
    # db version

    click.secho(
        "DONE database for {} {}{} at version {}".format(
            module,
            "initialized" if initialized else "exists",
            f" (flavor {dbflavor})" if dbflavor is not None else "",
            dbversion,
        ),
        fg="green",
        bold=True,
    )

    if flavor is not None and dbflavor != flavor:
        click.secho(
            f"WARNING requested flavor '{flavor}' != recorded flavor '{dbflavor}'",
            fg="red",
            bold=True,
        )


@db.command(name="version", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--all/--no-all",
    "show_all",
    help="Show version history.",
    default=False,
    show_default=True,
)
@click.option("--module-config-key", help="Module config key to lookup.", default=None)
@click.pass_context
def db_version(ctx, module, show_all, module_config_key=None):
    """Print the database version for the Software Heritage.

    Example::

        swh db version -d swh-test
        swh db version scheduler
        swh db version scrubber --module-config-key=scrubber_db

    """
    from swh.core.db.db_utils import get_database_info, import_swhmodule

    # use the db cnx from the config file; the expected config entry is either the given
    # module_config_key or defaulting to the module name (if module_config_key is not
    # provided)
    cfg = ctx.obj["config"].get(module_config_key or module, {})
    dbname = get_dburl_from_config(cfg)

    if not dbname:
        raise click.BadParameter(
            "Missing the postgresql connection configuration. Either fix your "
            "configuration file or use the --dbname option."
        )

    logger.debug("db_version dbname=%s", dbname)

    db_module, db_version, db_flavor = get_database_info(dbname)
    if db_module is None:
        click.secho(
            "WARNING the database does not have a dbmodule table.", fg="red", bold=True
        )
        db_module = module
    assert db_module == module, f"{db_module} (in the db) != {module} (given)"

    click.secho(f"module: {db_module}", fg="green", bold=True)

    if db_flavor is not None:
        click.secho(f"flavor: {db_flavor}", fg="green", bold=True)

    # instantiate the data source to retrieve the current (expected) db version
    datastore_factory = getattr(import_swhmodule(db_module), "get_datastore", None)
    if datastore_factory:
        datastore = datastore_factory(**cfg)
        code_version = datastore.current_version
        click.secho(
            f"current code version: {code_version}",
            fg="green" if code_version == db_version else "red",
            bold=True,
        )

    if not show_all:
        click.secho(f"version: {db_version}", fg="green", bold=True)
    else:
        from swh.core.db.db_utils import swh_db_versions

        versions = swh_db_versions(dbname)
        for version, tstamp, desc in versions:
            click.echo(f"{version} [{tstamp}] {desc}")


@db.command(name="upgrade", context_settings=CONTEXT_SETTINGS)
@click.argument("module", required=True)
@click.option(
    "--to-version",
    type=int,
    help="Upgrade up to version VERSION",
    metavar="VERSION",
    default=None,
)
@click.option(
    "--interactive/--non-interactive",
    help="Do not ask questions (use default answer to all questions)",
    default=True,
)
@click.option(
    "--module-config-key", help="Module configuration key to lookup.", default=None
)
@click.pass_context
def db_upgrade(ctx, module, to_version, interactive, module_config_key):
    """Upgrade the database for given module (to a given version if specified).

    Examples::

        swh db upgrade storage
        swh db upgrade scheduler --to-version=10
        swh db upgrade scrubber --to-version=10 --module-config-key=scrubber_db

    """
    from swh.core.db.db_utils import (
        get_database_info,
        import_swhmodule,
        swh_db_upgrade,
        swh_set_db_module,
    )

    # use the db cnx from the config file; the expected config entry is either the given
    # module_config_key or defaulting to the module name (if module_config_key is not
    # provided)
    cfg = ctx.obj["config"].get(module_config_key or module, {})
    dbname = get_dburl_from_config(cfg)

    if not dbname:
        raise click.BadParameter(
            "Missing the postgresql connection configuration. Either fix your "
            "configuration file or use the --dbname option."
        )

    logger.debug("db_version dbname=%s", dbname)

    db_module, db_version, db_flavor = get_database_info(dbname)
    if db_module is None:
        click.secho(
            "Warning: the database does not have a dbmodule table.",
            fg="yellow",
            bold=True,
        )
        if interactive and not click.confirm(
            f"Write the module information ({module}) in the database?", default=True
        ):
            raise click.BadParameter("Migration aborted.")
        swh_set_db_module(dbname, module)
        db_module = module

    if db_module != module:
        raise click.BadParameter(
            f"Error: the given module ({module}) does not match the value "
            f"stored in the database ({db_module})."
        )

    # instantiate the data source to retrieve the current (expected) db version
    datastore_factory = getattr(import_swhmodule(db_module), "get_datastore", None)
    if not datastore_factory:
        raise click.UsageError(
            "You cannot use this command on old-style datastore backend {db_module}"
        )
    datastore = datastore_factory(**cfg)
    ds_version = datastore.current_version
    if to_version is None:
        to_version = ds_version
    if to_version > ds_version:
        raise click.UsageError(
            f"The target version {to_version} is larger than the current version "
            f"{ds_version} of the datastore backend {db_module}"
        )

    if to_version == db_version:
        click.secho(
            f"No migration needed: the current version is {db_version}",
            fg="yellow",
        )
    else:
        new_db_version = swh_db_upgrade(dbname, module, to_version)
        click.secho(f"Migration to version {new_db_version} done", fg="green")
        if new_db_version < ds_version:
            click.secho(
                "Warning: migration was not complete: "
                f"the current version is {ds_version}",
                fg="yellow",
            )


def get_dburl_from_config(cfg):
    if cfg.get("cls") != "postgresql":
        raise click.BadParameter(
            "Configuration cls must be set to 'postgresql' for this command."
        )
    if "args" in cfg:
        # for bw compat
        cfg = cfg["args"]
    return cfg.get("db")



# Copyright (C) 2015-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import shutil
import stat
from subprocess import run
import tarfile
import zipfile

import magic

from . import utils


def _unpack_tar(tarpath: str, extract_dir: str) -> str:
    """Unpack tarballs unsupported by the standard python library. Examples
    include tar.Z, tar.lz, tar.x, etc....

    As this implementation relies on the `tar` command, this function supports
    the same compression the tar command supports.

    This expects the `extract_dir` to exist.

    Raises:
        shutil.ReadError in case of issue uncompressing the archive (tarpath
        does not exist, extract_dir does not exist, etc...)

    Returns:
        full path to the uncompressed directory.

    """
    try:
        run(["tar", "xf", tarpath, "-C", extract_dir], check=True)
        return extract_dir
    except Exception as e:
        raise shutil.ReadError(
            f"Unable to uncompress {tarpath} to {extract_dir}. Reason: {e}"
        )


def _unpack_zip(zippath: str, extract_dir: str) -> str:
    """Unpack zip files unsupported by the standard python library, for instance
    those with legacy compression type 6 (implode).

    This expects the `extract_dir` to exist.

    Raises:
        shutil.ReadError in case of issue uncompressing the archive (zippath
        does not exist, extract_dir does not exist, etc...)

    Returns:
        full path to the uncompressed directory.

    """
    try:
        run(["unzip", "-q", "-d", extract_dir, zippath], check=True)
        return extract_dir
    except Exception as e:
        raise shutil.ReadError(
            f"Unable to uncompress {zippath} to {extract_dir}. Reason: {e}"
        )


def _unpack_jar(jarpath: str, extract_dir: str) -> str:
    """Unpack jar files using standard Python module zipfile.

    This expects the `extract_dir` to exist.

    Raises:
        shutil.ReadError in case of issue uncompressing the archive (jarpath
        does not exist, extract_dir does not exist, etc...)

    Returns:
        full path to the uncompressed directory.

    """
    try:
        with zipfile.ZipFile(jarpath) as jar:
            jar.extractall(path=extract_dir)
        return extract_dir
    except Exception as e:
        raise shutil.ReadError(
            f"Unable to uncompress {jarpath} to {extract_dir}. Reason: {e}"
        )


def _unpack_zst(zstpath: str, extract_dir: str) -> str:
    """Unpack zst files unsupported by the standard python library. Example
    include tar.zst

    This expects the `extract_dir` to exist.

    Raises:
        shutil.ReadError in case of issue uncompressing the archive (zstpath
    """
    try:
        run(
            ["tar", "--force-local", "-I 'zstd'", "-xf", zstpath, "-C", extract_dir],
            check=True,
        )
        return extract_dir
    except Exception as e:
        raise shutil.ReadError(
            f"Unable to uncompress {zstpath} to {extract_dir}. Reason: {e}"
        )


def register_new_archive_formats():
    """Register new archive formats to uncompress"""
    registered_formats = [f[0] for f in shutil.get_unpack_formats()]
    for name, extensions, function in ADDITIONAL_ARCHIVE_FORMATS:
        if name in registered_formats:
            continue
        shutil.register_unpack_format(name, extensions, function)


_mime_to_archive_format = {
    "application/x-compress": "tar.Z|x",
    "application/x-tar": "tar",
    "application/x-bzip2": "bztar",
    "application/gzip": "gztar",
    "application/x-lzip": "tar.lz",
    "application/zip": "zip",
    "application/java-archive": "jar",
    "application/zstd": "tar.zst",
    "application/x-zstd": "tar.zst",
}


def uncompress(tarpath: str, dest: str):
    """Uncompress tarpath to dest folder if tarball is supported.

       Note that this fixes permissions after successfully
       uncompressing the archive.

    Args:
        tarpath: path to tarball to uncompress
        dest: the destination folder where to uncompress the tarball,
            it will be created if it does not exist

    Raises:
        ValueError when a problem occurs during unpacking

    """
    try:
        os.makedirs(dest, exist_ok=True)
        format = None
        # try to get archive format from extension
        for format_, exts, _ in shutil.get_unpack_formats():
            if any([tarpath.lower().endswith(ext.lower()) for ext in exts]):
                format = format_
                break
        # try to get archive format from file mimetype
        if format is None:
            m = magic.Magic(mime=True)
            mime = m.from_file(tarpath)
            format = _mime_to_archive_format.get(mime)
        shutil.unpack_archive(tarpath, extract_dir=dest, format=format)
    except shutil.ReadError as e:
        raise ValueError(f"Problem during unpacking {tarpath}. Reason: {e}")
    except NotImplementedError:
        if tarpath.lower().endswith(".zip") or format == "zip":
            _unpack_zip(tarpath, dest)
        else:
            raise

    normalize_permissions(dest)


def normalize_permissions(path: str):
    """Normalize the permissions of all files and directories under `path`.

    This makes all subdirectories and files with the user executable bit set mode
    0o0755, and all other files mode 0o0644.

    Args:
      path: the path under which permissions should be normalized
    """
    for dirpath, _, fnames in os.walk(path):
        os.chmod(dirpath, 0o0755)
        for fname in fnames:
            fpath = os.path.join(dirpath, fname)
            if not os.path.islink(fpath):
                is_executable = os.stat(fpath).st_mode & stat.S_IXUSR
                forced_mode = 0o0755 if is_executable else 0o0644
                os.chmod(fpath, forced_mode)


def _ls(rootdir):
    """Generator of filepath, filename from rootdir."""
    for dirpath, dirnames, fnames in os.walk(rootdir):
        for fname in dirnames + fnames:
            fpath = os.path.join(dirpath, fname)
            fname = utils.commonname(rootdir, fpath)
            yield fpath, fname


def _compress_zip(tarpath, files):
    """Compress dirpath's content as tarpath."""
    with zipfile.ZipFile(tarpath, "w") as z:
        for fpath, fname in files:
            z.write(fpath, arcname=fname)


def _compress_tar(tarpath, files):
    """Compress dirpath's content as tarpath."""
    with tarfile.open(tarpath, "w:bz2") as t:
        for fpath, fname in files:
            t.add(fpath, arcname=fname, recursive=False)


def compress(tarpath, nature, dirpath_or_files):
    """Create a tarball tarpath with nature nature.
    The content of the tarball is either dirpath's content (if representing
    a directory path) or dirpath's iterable contents.

    Compress the directory dirpath's content to a tarball.
    The tarball being dumped at tarpath.
    The nature of the tarball is determined by the nature argument.

    """
    if isinstance(dirpath_or_files, str):
        files = _ls(dirpath_or_files)
    else:  # iterable of 'filepath, filename'
        files = dirpath_or_files

    if nature == "zip":
        _compress_zip(tarpath, files)
    else:
        _compress_tar(tarpath, files)

    return tarpath


# Additional uncompression archive format support
ADDITIONAL_ARCHIVE_FORMATS = [
    # name, extensions, function
    ("tar.Z|x", [".tar.Z", ".tar.x"], _unpack_tar),
    ("jar", [".jar"], _unpack_jar),
    ("tbz2", [".tbz", "tbz2"], _unpack_tar),
    # FIXME: make this optional depending on the runtime lzip package install
    ("tar.lz", [".tar.lz"], _unpack_tar),
    ("crate", [".crate"], _unpack_tar),
    ("tar.zst", [".tar.zst", ".tar.zstd"], _unpack_zst),
]

register_new_archive_formats()



# Copyright (C) 2019-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib
import os
import shutil

import pytest

from swh.core import tarball


@pytest.fixture
def prepare_shutil_state():
    """Reset any shutil modification in its current state"""
    import shutil

    registered_formats = [f[0] for f in shutil.get_unpack_formats()]
    for format_id in tarball.ADDITIONAL_ARCHIVE_FORMATS:
        name = format_id[0]
        if name in registered_formats:
            shutil.unregister_unpack_format(name)

    return shutil


def test_compress_uncompress_zip(tmp_path):
    tocompress = tmp_path / "compressme"
    tocompress.mkdir()

    for i in range(10):
        fpath = tocompress / ("file%s.txt" % i)
        fpath.write_text("content of file %s" % i)

    zipfile = tmp_path / "archive.zip"
    tarball.compress(str(zipfile), "zip", str(tocompress))

    destdir = tmp_path / "destdir"
    tarball.uncompress(str(zipfile), str(destdir))

    lsdir = sorted(x.name for x in destdir.iterdir())
    assert ["file%s.txt" % i for i in range(10)] == lsdir


@pytest.mark.xfail(
    reason=(
        "Python's zipfile library doesn't support Info-ZIP's "
        "extension for file permissions."
    )
)
def test_compress_uncompress_zip_modes(tmp_path):
    tocompress = tmp_path / "compressme"
    tocompress.mkdir()

    fpath = tocompress / "text.txt"
    fpath.write_text("echo foo")
    fpath.chmod(0o644)

    fpath = tocompress / "executable.sh"
    fpath.write_text("echo foo")
    fpath.chmod(0o755)

    zipfile = tmp_path / "archive.zip"
    tarball.compress(str(zipfile), "zip", str(tocompress))

    destdir = tmp_path / "destdir"
    tarball.uncompress(str(zipfile), str(destdir))

    (executable_path, text_path) = sorted(destdir.iterdir())
    assert text_path.stat().st_mode == 0o100644  # succeeds, it's the default
    assert executable_path.stat().st_mode == 0o100755  # fails


def test_compress_uncompress_tar(tmp_path):
    tocompress = tmp_path / "compressme"
    tocompress.mkdir()

    for i in range(10):
        fpath = tocompress / ("file%s.txt" % i)
        fpath.write_text("content of file %s" % i)

    tarfile = tmp_path / "archive.tar"
    tarball.compress(str(tarfile), "tar", str(tocompress))

    destdir = tmp_path / "destdir"
    tarball.uncompress(str(tarfile), str(destdir))

    lsdir = sorted(x.name for x in destdir.iterdir())
    assert ["file%s.txt" % i for i in range(10)] == lsdir


def test_compress_uncompress_tar_modes(tmp_path):
    tocompress = tmp_path / "compressme"
    tocompress.mkdir()

    fpath = tocompress / "text.txt"
    fpath.write_text("echo foo")
    fpath.chmod(0o644)

    fpath = tocompress / "executable.sh"
    fpath.write_text("echo foo")
    fpath.chmod(0o755)

    tarfile = tmp_path / "archive.tar"
    tarball.compress(str(tarfile), "tar", str(tocompress))

    destdir = tmp_path / "destdir"
    tarball.uncompress(str(tarfile), str(destdir))

    (executable_path, text_path) = sorted(destdir.iterdir())
    assert text_path.stat().st_mode == 0o100644
    assert executable_path.stat().st_mode == 0o100755


def test_uncompress_tar_failure(tmp_path, datadir):
    """Unpack inexistent tarball should fail"""
    tarpath = os.path.join(datadir, "archives", "inexistent-archive.tar.Z")

    assert not os.path.exists(tarpath)

    with pytest.raises(ValueError, match="Problem during unpacking"):
        tarball.uncompress(tarpath, tmp_path)


def test_uncompress_tar(tmp_path, datadir):
    """Unpack supported tarball into an existent folder should be ok"""
    filename = "groff-1.02.tar.Z"
    tarpath = os.path.join(datadir, "archives", filename)

    assert os.path.exists(tarpath)

    extract_dir = os.path.join(tmp_path, filename)

    tarball.uncompress(tarpath, extract_dir)

    assert len(os.listdir(extract_dir)) > 0


def test_uncompress_tarzst(tmp_path, datadir):
    """Unpack supported zst tarball into an existent folder should be ok"""
    filename = "ca-certificates-20210603-1-any.pkg.tar.zst"
    tarpath = os.path.join(datadir, "archives", filename)

    assert os.path.exists(tarpath)

    extract_dir = os.path.join(tmp_path, filename)

    tarball.uncompress(tarpath, extract_dir)

    assert len(os.listdir(extract_dir)) > 0


def test_register_new_archive_formats(prepare_shutil_state):
    """Registering new archive formats should be fine"""
    unpack_formats_v1 = [f[0] for f in shutil.get_unpack_formats()]
    for format_id in tarball.ADDITIONAL_ARCHIVE_FORMATS:
        assert format_id[0] not in unpack_formats_v1

    # when
    tarball.register_new_archive_formats()

    # then
    unpack_formats_v2 = [f[0] for f in shutil.get_unpack_formats()]
    for format_id in tarball.ADDITIONAL_ARCHIVE_FORMATS:
        assert format_id[0] in unpack_formats_v2


def test_uncompress_archives(tmp_path, datadir):
    """High level call uncompression on supported archives"""
    archive_dir = os.path.join(datadir, "archives")
    archive_files = os.listdir(archive_dir)

    for archive_file in archive_files:
        archive_path = os.path.join(archive_dir, archive_file)
        extract_dir = os.path.join(tmp_path, archive_file)
        tarball.uncompress(archive_path, dest=extract_dir)
        assert len(os.listdir(extract_dir)) > 0


def test_normalize_permissions(tmp_path):
    for perms in range(0o1000):
        filename = str(perms)
        file_path = tmp_path / filename
        file_path.touch()
        file_path.chmod(perms)

    for file in tmp_path.iterdir():
        assert file.stat().st_mode == 0o100000 | int(file.name)

    tarball.normalize_permissions(str(tmp_path))

    for file in tmp_path.iterdir():
        if int(file.name) & 0o100:  # original file was executable for its owner
            assert file.stat().st_mode == 0o100755
        else:
            assert file.stat().st_mode == 0o100644


def test_unpcompress_zip_imploded(tmp_path, datadir):
    """Unpack a zip archive with compression type 6 (implode),
    not supported by python zipfile module.

    """
    filename = "msk316src.zip"
    zippath = os.path.join(datadir, "archives", filename)

    assert os.path.exists(zippath)

    extract_dir = os.path.join(tmp_path, filename)

    tarball.uncompress(zippath, extract_dir)

    assert len(os.listdir(extract_dir)) > 0


def test_uncompress_upper_archive_extension(tmp_path, datadir):
    """Copy test archives in a temporary directory but turn their names
    to uppercase, then check they can be successfully extracted.
    """
    archives_path = os.path.join(datadir, "archives")
    archive_files = [
        f
        for f in os.listdir(archives_path)
        if os.path.isfile(os.path.join(archives_path, f))
    ]
    for archive_file in archive_files:
        archive_file_upper = os.path.join(tmp_path, archive_file.upper())
        extract_dir = os.path.join(tmp_path, archive_file)
        shutil.copy(os.path.join(archives_path, archive_file), archive_file_upper)
        tarball.uncompress(archive_file_upper, extract_dir)
        assert len(os.listdir(extract_dir)) > 0


def test_uncompress_archive_no_extension(tmp_path, datadir):
    """Copy test archives in a temporary directory but turn their names
    to their md5 sums, then check they can be successfully extracted.
    """
    archives_path = os.path.join(datadir, "archives")
    archive_files = [
        f
        for f in os.listdir(archives_path)
        if os.path.isfile(os.path.join(archives_path, f))
    ]
    for archive_file in archive_files:
        archive_file_path = os.path.join(archives_path, archive_file)
        with open(archive_file_path, "rb") as f:
            md5sum = hashlib.md5(f.read()).hexdigest()
        archive_file_md5sum = os.path.join(tmp_path, md5sum)
        extract_dir = os.path.join(tmp_path, archive_file)
        shutil.copy(archive_file_path, archive_file_md5sum)
        tarball.uncompress(archive_file_md5sum, extract_dir)
        assert len(os.listdir(extract_dir)) > 0



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Default values for gunicorn's configuration.

Other packages may override them by importing `*` from this module
and redefining functions and variables they want.

May be imported by gunicorn using
`--config 'python:swh.core.api.gunicorn_config'`."""


from ..sentry import init_sentry


def post_fork(
    server,
    worker,
    *,
    default_sentry_dsn=None,
    flask=True,
    sentry_integrations=None,
    extra_sentry_kwargs={},
    disable_logging_events=True,
):
    # Initializes sentry as soon as possible in gunicorn's worker processes.

    sentry_integrations = sentry_integrations or []
    if flask:
        from sentry_sdk.integrations.flask import FlaskIntegration

        sentry_integrations.append(FlaskIntegration())

    init_sentry(
        sentry_dsn=default_sentry_dsn,
        integrations=sentry_integrations,
        extra_kwargs=extra_sentry_kwargs,
        disable_logging_events=disable_logging_events,
    )
